from distutils.version import LooseVersion
import glob
import os
import warnings
from multiprocessing import Pool
import dpgen.auto_test.lib.util as util
from dpgen import dlog
from dpgen.util import sepline
from dpgen.auto_test.EOS import EOS
from dpgen.auto_test.Elastic import Elastic
from dpgen.auto_test.Interstitial import Interstitial
from dpgen.auto_test.Surface import Surface
from dpgen.auto_test.Vacancy import Vacancy
from dpgen.auto_test.calculator import make_calculator
from dpgen.dispatcher.Dispatcher import make_dispatcher
from dpgen.dispatcher.Dispatcher import make_submission
from dpgen.remote.decide_machine import convert_mdata
from dpgen.auto_test.lib.utils import create_path
lammps_task_type = ['deepmd', 'meam', 'eam_fs', 'eam_alloy']


def make_property_instance(paramters):
    """
    Make an instance of Property
    """
    prop_type = paramters['type']
    if prop_type == 'eos':
        return EOS(paramters)
    elif prop_type == 'elastic':
        return Elastic(paramters)
    elif prop_type == 'vacancy':
        return Vacancy(paramters)
    elif prop_type == 'interstitial':
        return Interstitial(paramters)
    elif prop_type == 'surface':
        return Surface(paramters)
    else:
        raise RuntimeError(f'unknown property type {prop_type}')


def make_property(confs,
                  inter_param,
                  property_list):
    # find all POSCARs and their name like mp-xxx
    # ...
    # conf_dirs = glob.glob(confs)
    # conf_dirs.sort()
    conf_dirs = []
    for conf in confs:
        conf_dirs.extend(glob.glob(conf))
    conf_dirs.sort()
    for ii in conf_dirs:
        sepline(ch=ii, screen=True)
        for jj in property_list:
            if jj.get("skip", False):
                continue
            if 'init_from_suffix' and 'output_suffix' in jj:
                do_refine = True
                suffix = jj['output_suffix']
            elif 'reproduce' in jj and jj['reproduce']:
                do_refine = False
                suffix = 'reprod'
            else:
                do_refine = False
                suffix = '00'
            # generate working directory like mp-xxx/eos_00 if jj['type'] == 'eos'
            # handel the exception that the working directory exists
            # ...

            # determine the suffix: from scratch or refine
            # ...

            property_type = jj['type']
            path_to_equi = os.path.join(ii, 'relaxation', 'relax_task')
            path_to_work = os.path.join(ii, property_type + '_' + suffix)

            create_path(path_to_work)

            prop = make_property_instance(jj)
            task_list = prop.make_confs(path_to_work, path_to_equi, do_refine)

            inter_param_prop = inter_param
            if 'cal_setting' in jj and 'overwrite_interaction' in jj['cal_setting']:
                inter_param_prop = jj['cal_setting']['overwrite_interaction']

            for kk in task_list:
                poscar = os.path.join(kk, 'POSCAR')
                inter = make_calculator(inter_param_prop, poscar)
                inter.make_potential_files(kk)
                dlog.debug(prop.task_type())  ### debug
                inter.make_input_file(kk, prop.task_type(), prop.task_param())

            prop.post_process(task_list)  # generate same KPOINTS file for elastic when doing VASP


def run_property(confs,
                 inter_param,
                 property_list,
                 mdata):
    # find all POSCARs and their name like mp-xxx
    # ...
    # conf_dirs = glob.glob(confs)
    # conf_dirs.sort()
    processes = len(property_list)
    pool = Pool(processes=processes)
    print("Submit job via %d processes" % processes)
    conf_dirs = []
    for conf in confs:
        conf_dirs.extend(glob.glob(conf))
    conf_dirs.sort()
    task_list = []
    work_path_list = []
    multiple_ret = []
    for ii in conf_dirs:
        sepline(ch=ii, screen=True)
        for jj in property_list:
            # determine the suffix: from scratch or refine
            # ...
            if jj.get("skip", False):
                continue
            if 'init_from_suffix' and 'output_suffix' in jj:
                suffix = jj['output_suffix']
            elif 'reproduce' in jj and jj['reproduce']:
                suffix = 'reprod'
            else:
                suffix = '00'

            property_type = jj['type']
            path_to_work = os.path.abspath(os.path.join(ii, property_type + '_' + suffix))

            work_path_list.append(path_to_work)
            tmp_task_list = glob.glob(os.path.join(path_to_work, 'task.[0-9]*[0-9]'))
            tmp_task_list.sort()
            task_list.append(tmp_task_list)

            inter_param_prop = inter_param
            if 'cal_setting' in jj and 'overwrite_interaction' in jj['cal_setting']:
                inter_param_prop = jj['cal_setting']['overwrite_interaction']

            # dispatch the tasks
            # POSCAR here is useless
            virtual_calculator = make_calculator(inter_param_prop, "POSCAR")
            forward_files = virtual_calculator.forward_files(property_type)
            forward_common_files = virtual_calculator.forward_common_files(property_type)
            backward_files = virtual_calculator.backward_files(property_type)
            #    backward_files += logs
            # ...
            inter_type = inter_param_prop['type']
            # vasp
            if inter_type == "vasp":
                mdata = convert_mdata(mdata, ["fp"])
            elif inter_type in lammps_task_type:
                mdata = convert_mdata(mdata, ["model_devi"])
            else:
                raise RuntimeError("unknown task %s, something wrong" % inter_type)

            work_path = path_to_work
            all_task = tmp_task_list
            run_tasks = util.collect_task(all_task, inter_type)
            if len(run_tasks) == 0:
                continue
            else:
                ret = pool.apply_async(worker, (work_path,
                                                all_task,
                                                forward_common_files,
                                                forward_files,
                                                backward_files,
                                                mdata,
                                                inter_type,
                                                ))
                multiple_ret.append(ret)
    pool.close()
    pool.join()
    for ii in range(len(multiple_ret)):
        if not multiple_ret[ii].successful():
            print("ERROR:", multiple_ret[ii].get())
            raise RuntimeError("Job %d is not successful!" % ii)
    print('%d jobs are finished' % len(multiple_ret))


def worker(work_path,
           all_task,
           forward_common_files,
           forward_files,
           backward_files,
           mdata,
           inter_type):
    run_tasks = [os.path.basename(ii) for ii in all_task]
    machine, resources, command, group_size = util.get_machine_info(mdata, inter_type)
    api_version = mdata.get('api_version', '0.9')
    if LooseVersion(api_version) < LooseVersion('1.0'):
        warnings.warn(f"the dpdispatcher will be updated to new version."
            f"And the interface may be changed. Please check the documents for more details")
        disp = make_dispatcher(machine, resources, work_path, run_tasks, group_size)
        disp.run_jobs(resources,
                  command,
                  work_path,
                  run_tasks,
                  group_size,
                  forward_common_files,
                  forward_files,
                  backward_files,
                  outlog='outlog',
                  errlog='errlog')
    elif LooseVersion(api_version) >= LooseVersion('1.0'):
        submission = make_submission(
                mdata_machine=machine,
                mdata_resources=resources,
                commands=[command],
                work_path=work_path,
                run_tasks=run_tasks,
                group_size=group_size,
                forward_common_files=forward_common_files,
                forward_files=forward_files,
                backward_files=backward_files,
                outlog = 'outlog',
                errlog = 'errlog'
            )
        submission.run_submission()

def post_property(confs,
                  #                  inter_param,
                  property_list):
    # find all POSCARs and their name like mp-xxx
    # ...
    #    task_list = []
    # conf_dirs = glob.glob(confs)
    # conf_dirs.sort()
    conf_dirs = []
    for conf in confs:
        conf_dirs.extend(glob.glob(conf))
    conf_dirs.sort()
    for ii in conf_dirs:
        for jj in property_list:
            # determine the suffix: from scratch or refine
            # ...
            if jj.get("skip", False):
                continue
            if 'init_from_suffix' and 'output_suffix' in jj:
                suffix = jj['output_suffix']
            elif 'reproduce' in jj and jj['reproduce']:
                suffix = 'reprod'
            else:
                suffix = '00'
            property_type = jj['type']
            path_to_work = os.path.join(ii, property_type + '_' + suffix)
            prop = make_property_instance(jj)
            prop.compute(os.path.join(path_to_work, 'result.json'), os.path.join(path_to_work, 'result.out'),
                         path_to_work)



#!/data1/anguse/zijian/deepmd-kit-devel/virtual_tf_2/bin/python
import deepmd.DeepPot as DP
import dpdata
import numpy as np
import os, sys
import json
def calc_model_devi_f(fs):
    '''
        fs : numpy.ndarray, size of `n_models x n_frames x n_atoms x 3`
    '''
    fs_mean = np.mean(fs, axis=0)
    # print(fs_mean.shape)
    fs_err = np.sum((fs - fs_mean) ** 2, axis=-1)
    # print(fs_err.shape)
    fs_devi = np.mean(fs_err, axis=0) ** 0.5
    # print(fs_devi.shape) 
    max_devi_f = np.max(fs_devi, axis=1)
    # min_devi_f = np.min(fs_devi, axis=1)
    # avg_devi_f = np.mean(fs_devi, axis=1)
    return max_devi_f

def write_model_devi_out(system, models, fname=None, trj_freq = 10):
    forces = []
    for model in models:
        labeled = system.predict(model)
        forces.append(labeled['forces'])
    forces = np.array(forces)
    max_devi_f = calc_model_devi_f(forces)
    model_devi_out = np.zeros((system.get_nframes(), 7))
    model_devi_out[:, 0] += np.arange(system.get_nframes()) * trj_freq
    model_devi_out[:, 4] += max_devi_f
    if fname is not None:
        np.savetxt(fname,
                   model_devi_out,
                   fmt=['%d'] + ['%.8e' for _ in range(6)],
                   delimiter='\t',
                   header='step\tmax_devi_e\tmin_devi_e\tavg_devi_e\tmax_devi_f\tmin_devi_f\tavg_devi_f')
    return model_devi_out

if __name__ == "__main__":
    system = dpdata.System(sys.argv[1], fmt='gromacs/gro')
    if os.path.isfile("job.json"):
        trj_freq = json.load(open("job.json")).get("trj_freq", 10)
    else:
        trj_freq = 10
    if not os.path.isdir("traj"):
        os.mkdir("traj")
    for i in range(system.get_nframes()):
        system[i].to_gromacs_gro("traj/%d.gromacstrj" % (trj_freq * i) )
    models = [DP(f"../graph.{ii:03}.pb") for ii in range(4)]
    write_model_devi_out(system, models, "model_devi.out", trj_freq)



import numpy as np
from dpdata.abacus.scf import get_cell, get_coords
from dpgen.auto_test.lib.vasp import reciprocal_box
import os
bohr2ang = 0.52917721067
def make_abacus_scf_kpt(fp_params):
    # Make KPT file for abacus pw scf calculation.
    # KPT file is the file containing k points infomation in ABACUS scf calculation.
    k_points = [1, 1, 1, 0, 0, 0]
    if "k_points" in fp_params:
        k_points = fp_params["k_points"]
        if len(k_points) != 6:
            raise RuntimeError("k_points has to be a list containig 6 integers specifying MP k points generation.")
    ret = "K_POINTS\n0\nGamma\n"
    for i in range(6):
        ret += str(k_points[i]) + " "
    return ret

def make_abacus_scf_input(fp_params):
    # Make INPUT file for abacus pw scf calculation.
    ret = "INPUT_PARAMETERS\n"
    ret += "calculation scf\n"
    for key in fp_params:
        if key == 'ntype':
            assert(fp_params['ntype'] >= 0 and type(fp_params["ntype"]) == int),  "'ntype' should be a positive integer."
            ret += "ntype %d\n" % fp_params['ntype']
        #ret += "pseudo_dir ./\n"
        elif key == "ecutwfc":
            assert(fp_params["ecutwfc"] >= 0) ,  "'ntype' should be non-negative."
            ret += "ecutwfc %f\n" % fp_params["ecutwfc"]
        elif key == "dr2":
            ret += "dr2 %e\n" % fp_params["dr2"]
        elif key == "niter":
            assert(fp_params['niter'] >= 0 and type(fp_params["niter"])== int), "'niter' should be a positive integer."
            ret += "niter %d\n" % fp_params["niter"]    
        elif key == "basis_type":
            assert(fp_params["basis_type"] in ["pw", "lcao", "lcao_in_pw"]) , "'basis_type' must in 'pw', 'lcao' or 'lcao_in_pw'."
            ret+= "basis_type %s\n" % fp_params["basis_type"]
        elif key == "dft_functional":
            ret += "dft_functional %s\n" % fp_params["dft_functional"]
        elif key == "gamma_only":
            #assert(fp_params["gamma_only"] ==1 ) , "'gamma_only' should be 1. Multi-k algorithm will be supported after the KPT generator is completed."
            ret+= "gamma_only %d\n" % fp_params["gamma_only"]  
        elif key == "mixing_type":
            assert(fp_params["mixing_type"] in ["plain", "kerker", "pulay", "pulay-kerker", "broyden"])
            ret += "mixing_type %s\n" % fp_params["mixing_type"]
        elif key == "mixing_beta":
            assert(fp_params["mixing_beta"] >= 0 and fp_params["mixing_beta"] < 1), "'mixing_beta' should between 0 and 1."
            ret += "mixing_beta %f\n" % fp_params["mixing_beta"]
        elif key == "symmetry":
            assert(fp_params["symmetry"] == 0 or fp_params["symmetry"] == 1), "'symmetry' should be either 0 or 1."
            ret += "symmetry %d\n" % fp_params["symmetry"]
        elif key == "nbands":
            assert(fp_params["nbands"] > 0 and type(fp_params["nbands"]) == int), "'nbands' should be a positive integer."
            ret += "nbands %d\n" % fp_params["nbands"]
        elif key == "nspin":
            assert(fp_params["nspin"] == 1 or fp_params["nspin"] == 2 or fp_params["nspin"] == 4), "'nspin' can anly take 1, 2 or 4"
            ret += "nspin %d\n" % fp_params["nspin"]
        elif key == "ks_solver":
            assert(fp_params["ks_solver"] in ["cg", "dav", "lapack", "genelpa", "hpseps", "scalapack_gvx"]), "'ks_sover' should in 'cgx', 'dav', 'lapack', 'genelpa', 'hpseps', 'scalapack_gvx'."
            ret += "ks_solver %s\n" % fp_params["ks_solver"]
        elif key == "smearing":
            assert(fp_params["smearing"] in ["gaussian", "fd", "fixed", "mp", "mp2", "mv"]), "'smearing' should in 'gaussian', 'fd', 'fixed', 'mp', 'mp2', 'mv'. "
            ret += "smearing %s\n" % fp_params["smearing"]
        elif key == "sigma":
            assert(fp_params["sigma"] >= 0), "'sigma' should be non-negative."
            ret += "sigma %f\n" % fp_params["sigma"]
        elif key == "force":
            assert(fp_params["force"] == 0  or fp_params["force"] == 1), "'force' should be either 0 or 1."
            ret += "force %d\n" % fp_params["force"]
        elif key == "stress":
            assert(fp_params["stress"] == 0  or fp_params["stress"] == 1), "'stress' should be either 0 or 1."
            ret += "stress %d\n" % fp_params["stress"]    
        #paras for deepks
        elif key == "deepks_out_labels":
            assert(fp_params["deepks_out_labels"] == 0 or fp_params["deepks_out_labels"] == 1), "'deepks_out_labels' should be either 0 or 1."
            ret += "deepks_out_labels %d\n" % fp_params["deepks_out_labels"]
        elif key == "deepks_descriptor_lmax":
            assert(fp_params["deepks_descriptor_lmax"] >= 0),  "'deepks_descriptor_lmax' should be  a positive integer."
            ret += "deepks_descriptor_lmax %d\n" % fp_params["deepks_descriptor_lmax"]
        elif key == "deepks_scf":
            assert(fp_params["deepks_scf"] == 0  or fp_params["deepks_scf"] == 1), "'deepks_scf' should be either 0 or 1."
            ret += "deepks_scf %d\n" % fp_params["deepks_scf"]
        elif key == "deepks_model":
            ret += "deepks_model %s\n" % fp_params["deepks_model"]
        elif key != "k_points": # "k_points key is used to generate KPT file."
            ret += "%s %s\n" % (key, str(fp_params[key]))
    return ret

def make_abacus_scf_stru(sys_data, fp_pp_files, fp_orb_files = None, fp_dpks_descriptor = None, fp_params = None):
    atom_names = sys_data['atom_names']
    atom_numbs = sys_data['atom_numbs']
    assert(len(atom_names) == len(fp_pp_files)), "the number of pp_files must be equal to the number of atom types. "
    assert(len(atom_names) == len(atom_numbs)), "Please check the name of atoms. "
    cell = sys_data["cells"].reshape([3, 3])
    coord = sys_data['coords'].reshape([sum(atom_numbs), 3])
    #volume = np.linalg.det(cell)
    #lattice_const = np.power(volume, 1/3)

    ret = "ATOMIC_SPECIES\n"
    for iatom in range(len(atom_names)):
        if 'atom_masses' not in sys_data:
            ret += atom_names[iatom] + " 1.00 " + fp_pp_files[iatom] + "\n"
        else:
            ret += atom_names[iatom] + " %.3f "%sys_data['atom_masses'][iatom] + fp_pp_files[iatom] + "\n"

    if fp_params is not None and "lattice_constant" in fp_params:
        ret += "\nLATTICE_CONSTANT\n"
        ret += str(fp_params["lattice_constant"]) + "\n\n" # in Bohr, in this way coord and cell are in Angstrom 
    else:
        ret += "\nLATTICE_CONSTANT\n"
        ret += str(1/bohr2ang) + "\n\n"

    ret += "LATTICE_VECTORS\n"
    for ix in range(3):
        for iy in range(3):
            ret += str(cell[ix][iy]) + " "
        ret += "\n"
    ret += "\n"

    ret += "ATOMIC_POSITIONS\n"
    ret += "Cartesian    # Cartesian(Unit is LATTICE_CONSTANT)\n"
    #ret += "\n"
    natom_tot = 0
    for iele in range(len(atom_names)):
        ret += atom_names[iele] + "\n"
        ret += "0.0\n"
        ret += str(atom_numbs[iele]) + "\n"
        for iatom in range(atom_numbs[iele]):
            ret += "%.12f %.12f %.12f %d %d %d\n" % (coord[natom_tot, 0], coord[natom_tot, 1], coord[natom_tot, 2], 1, 1, 1)
            natom_tot += 1
    assert(natom_tot == sum(atom_numbs))

    if fp_orb_files is not None:
        ret +="\nNUMERICAL_ORBITAL\n"
        assert(len(fp_orb_files)==len(atom_names))
        for iatom in range(len(atom_names)):
            ret += fp_orb_files[iatom] +"\n"

    if fp_dpks_descriptor is not None:
        ret +="\nNUMERICAL_DESCRIPTOR\n"
        ret +="%s\n"%fp_dpks_descriptor

    return ret

def get_abacus_input_parameters(INPUT):
    with open(INPUT) as fp:
        inlines = fp.read().split("\n")
    input_parameters = {}
    for line in inlines:
        if line.split() == [] or len(line.split()) < 2 :
            continue
        parameter_name = line.split()[0]
        parameter_value = line.split()[1]
        input_parameters[parameter_name] = parameter_value
    return input_parameters

def get_mass_from_STRU(geometry_inlines, inlines, atom_names):
    nele = None
    for line in inlines:
        if line.split() == []:
            continue
        if "ntype" in line and "ntype" == line.split()[0]:
            nele = int(line.split()[1])
    assert(nele is not None)
    mass_list = [0 for i in atom_names]
    pp_file_list = [i for i in atom_names]
    for iline, line in enumerate(geometry_inlines):
        if line.split() == []:
            continue
        if "ATOMIC_SPECIES" == line.split()[0]:
            for iele1 in range(1, 1+nele):
                for iele2 in range(nele):
                    if geometry_inlines[iline+iele1].split()[0] == atom_names[iele2]:
                        mass_list[iele2] = float(geometry_inlines[iline+iele1].split()[1])
                        pp_file_list[iele2] = geometry_inlines[iline+iele1].split()[2]
    for iele in range(len(mass_list)):
        assert(mass_list[iele] > 0)
    return mass_list, pp_file_list

def get_natoms_from_stru(geometry_inlines):
    key_words_list = ["ATOMIC_SPECIES", "NUMERICAL_ORBITAL", "LATTICE_CONSTANT", "LATTICE_VECTORS", "ATOMIC_POSITIONS"]
    keyword_sequence = []
    keyword_line_index = []
    atom_names = []
    atom_numbs = []
    for iline, line in enumerate(geometry_inlines):
        if line.split() == []:
            continue
        have_key_word = False
        for keyword in key_words_list:
            if keyword in line and keyword == line.split()[0]:
                keyword_sequence.append(keyword)
                keyword_line_index.append(iline)
    assert(len(keyword_line_index) == len(keyword_sequence))
    assert(len(keyword_sequence) > 0)
    keyword_line_index.append(len(geometry_inlines))
    for idx, keyword in enumerate(keyword_sequence):
        if keyword == "ATOMIC_POSITIONS":
            iline = keyword_line_index[idx]+2
            while iline < keyword_line_index[idx+1]-1:
                atom_names.append(geometry_inlines[iline].split()[0])
                atom_numbs.append(int(geometry_inlines[iline+2].split()[0]))
                iline += 3+atom_numbs[-1]
    return atom_names, atom_numbs

def get_additional_from_STRU(geometry_inlines, nele):
    dpks_descriptor_kw = "NUMERICAL_DESCRIPTOR"
    orb_file_kw = "NUMERICAL_ORBITAL"
    dpks_descriptor = None
    orb_file = None
    for iline in range(len(geometry_inlines)):
        if len(geometry_inlines[iline]) > 0:
            if orb_file_kw == geometry_inlines[iline].split()[0]:
                orb_file = []
                for iele in range(nele):
                    orb_file.append(geometry_inlines[iline + iele + 1].rstrip())
            if dpks_descriptor_kw == geometry_inlines[iline].split()[0]:
                dpks_descriptor = geometry_inlines[iline + 1].rstrip()
    return orb_file, dpks_descriptor

def get_abacus_STRU(STRU, INPUT = None, n_ele = None):
    # read in geometry from STRU file. n_ele is the number of elements.
    # Either n_ele or INPUT should be provided.
    with open(STRU, 'r') as fp:
        geometry_inlines = fp.read().split('\n')
    for iline, line in enumerate(geometry_inlines):
        if line.split() == [] or len(line) == 0:
            del geometry_inlines[iline]
    geometry_inlines.append("")
    celldm, cell = get_cell(geometry_inlines) 
    if n_ele is None and INPUT is not None:
        assert(os.path.isfile(INPUT)), "file %s should exists" % INPUT
        with open(INPUT, 'r') as fp:
            inlines = fp.read().split('\n')
        atom_names, natoms, types, coords = get_coords(celldm, cell, geometry_inlines, inlines) 
    elif n_ele is not None and INPUT is None:
        assert(n_ele > 0)
        inlines = ["ntype %d" %n_ele]
        atom_names, natoms, types, coords = get_coords(celldm, cell, geometry_inlines, inlines)
    else:
        atom_names, atom_numbs = get_natoms_from_stru(geometry_inlines)
        inlines = ["ntype %d" %len(atom_numbs)]
        atom_names, natoms, types, coords = get_coords(celldm, cell, geometry_inlines, inlines)
    masses, pp_files = get_mass_from_STRU(geometry_inlines, inlines, atom_names)
    orb_files, dpks_descriptor = get_additional_from_STRU(geometry_inlines, len(masses))
    data = {}
    data['atom_names'] = atom_names
    data['atom_numbs'] = natoms
    data['atom_types'] = types
    data['cells'] = cell
    data['coords'] = coords
    data['atom_masses'] = masses # Notice that this key is not defined in dpdata system. 
    data['pp_files'] = pp_files
    data['orb_files'] = orb_files
    data['dpks_descriptor'] = dpks_descriptor
    return data

def make_supercell_abacus(from_struct, super_cell):
    if "types" in from_struct:
        from_struct["types"] = from_struct["types"] * super_cell[0] * super_cell[1] * super_cell[2]
    for ix in range(super_cell[0]):
        for iy in range(super_cell[1]):
            for iz in range(super_cell[2]):
                if ix == 0 and iy == 0 and iz == 0:
                    continue
                for ia in range(sum(from_struct["atom_numbs"])):
                    coord = from_struct['coords'][ia] + from_struct['cells'][0]*ix + from_struct['cells'][1]*iy + from_struct['cells'][2]*iz
                    from_struct['coords'] = np.vstack([from_struct['coords'], coord])
    from_struct["atom_numbs"] = [i * super_cell[0] * super_cell[1] * super_cell[2] for i in from_struct["atom_numbs"]]
    from_struct['cells'][0] *= super_cell[0]
    from_struct['cells'][1] *= super_cell[1]
    from_struct['cells'][2] *= super_cell[2]
    return from_struct

def make_kspacing_kpoints_stru(stru, kspacing) :
    # adapted from dpgen.autotest.lib.vasp.make_kspacing_kpoints
    if type(kspacing) is not list:
        kspacing = [kspacing, kspacing, kspacing]
    box = stru['cells']
    rbox = reciprocal_box(box)
    kpoints = [max(1,(np.ceil(2 * np.pi * np.linalg.norm(ii) / ks).astype(int))) for ii,ks in zip(rbox,kspacing)]
    kpoints += [0, 0, 0]
    return kpoints

if __name__ == "__main__":
    fp_params = {"k_points": [1, 1, 1, 0, 0, 0]}
    ret = make_abacus_scf_kpt(fp_params)
    print(ret)


import os, sys, json, glob, shutil
import dpdata
import numpy as np
import unittest
import dpdata
from monty.serialization import loadfn, dumpfn
from pymatgen.core import Structure
from pymatgen.io.vasp import Incar
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.analysis.defects.core import Interstitial as pmg_Interstitial

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__package__ = 'auto_test'

from .context import make_kspacing_kpoints
from .context import setUpModule

from dpgen.auto_test.Interstitial import Interstitial


class TestInterstitial(unittest.TestCase):

    def setUp(self):
        _jdata = {
            "structures": ["confs/std-bcc"],
            "interaction": {
                "type": "vasp",
                "incar": "vasp_input/INCAR.rlx",
                "potcar_prefix": "vasp_input",
                "potcars": {"V": "POTCAR"}
            },
            "properties": [
                {
                    "type": "interstitial",
                    "supercell": [1, 1, 1],
                    "insert_ele": ["V"],
                    "bcc_self": True
                }
            ]
        }

        self.equi_path = 'confs/std-bcc/relaxation/relax_task'
        self.source_path = 'equi/vasp'
        self.target_path = 'confs/std-bcc/interstitial_00'
        if not os.path.exists(self.equi_path):
            os.makedirs(self.equi_path)
        if not os.path.exists(self.target_path):
            os.makedirs(self.target_path)

        self.confs = _jdata["structures"]
        self.inter_param = _jdata["interaction"]
        self.prop_param = _jdata['properties']

        self.interstitial = Interstitial(_jdata['properties'][0])

    def tearDown(self):
        if os.path.exists(self.equi_path):
            shutil.rmtree(self.equi_path)
        if os.path.exists(self.target_path):
            shutil.rmtree(self.target_path)

    def test_task_type(self):
        self.assertEqual('interstitial', self.interstitial.task_type())

    def test_task_param(self):
        self.assertEqual(self.prop_param[0], self.interstitial.task_param())

    def test_make_confs_bcc(self):
        if not os.path.exists(os.path.join(self.equi_path, 'CONTCAR')):
            with self.assertRaises(RuntimeError):
                self.interstitial.make_confs(self.target_path, self.equi_path)
        shutil.copy(os.path.join(self.source_path, 'CONTCAR_V_bcc'), os.path.join(self.equi_path, 'CONTCAR'))
        task_list = self.interstitial.make_confs(self.target_path, self.equi_path)
        dfm_dirs = glob.glob(os.path.join(self.target_path, 'task.*'))
        self.assertEqual(len(dfm_dirs), 7)

        incar0 = Incar.from_file(os.path.join('vasp_input', 'INCAR.rlx'))
        incar0['ISIF'] = 3

        self.assertEqual(os.path.realpath(os.path.join(self.equi_path, 'CONTCAR')),
                         os.path.realpath(os.path.join(self.target_path, 'POSCAR')))
        ref_st = Structure.from_file(os.path.join(self.target_path, 'POSCAR'))
        dfm_dirs.sort()
        for ii in dfm_dirs[:4]:
            st_file = os.path.join(ii, 'POSCAR')
            self.assertTrue(os.path.isfile(st_file))
            st0 = Structure.from_file(st_file)
            inter_site = st0[-1]
            inter = pmg_Interstitial(ref_st, inter_site, charge=0.0)
            st1 = inter.generate_defect_structure(self.prop_param[0]['supercell'])
            self.assertEqual(st0, st1)

        for ii in dfm_dirs[4:]:
            st_file = os.path.join(ii, 'POSCAR')
            self.assertTrue(os.path.isfile(st_file))
            st0 = Structure.from_file(st_file)
            inter_site1 = st0.pop(-1)
            inter_site2 = st0.pop(-1)
            center = (inter_site1.coords + inter_site2.coords) / 2
            self.assertTrue((center[0] - center[1]) < 1e-4)
            self.assertTrue((center[1] - center[2]) < 1e-4)



from typing import Tuple

from dargs import Argument

from dpgen.dispatcher.Dispatcher import mdata_arginfo


def general_mdata_arginfo(name: str, tasks: Tuple[str]) -> Argument:
    """Generate arginfo for general mdata.

    Parameters
    ----------
    name : str
        mdata name
    tasks : tuple[str]
        tuple of task keys, e.g. ("train", "model_devi", "fp")
    
    Returns
    -------
    Argument
        arginfo
    """
    
    doc_api_version = "Please set to 1.0"
    doc_run_mdata = "machine.json file"
    arg_api_version = Argument("api_version", str, optional=False, doc=doc_api_version)

    sub_fields = [arg_api_version]
    doc_mdata = "Parameters of command, machine, and resources for %s"
    for task in tasks:
        sub_fields.append(Argument(
            task, dict, optional=False, sub_fields=mdata_arginfo(),
            doc=doc_mdata % task,
        ))
    return Argument(name, dict, sub_fields=sub_fields, doc=doc_run_mdata)



import glob
import json
import os
import re

from monty.serialization import loadfn, dumpfn
from pymatgen.analysis.defects.generators import InterstitialGenerator
from pymatgen.core.structure import Structure

import dpgen.auto_test.lib.lammps as lammps
from dpgen.auto_test.Property import Property
from dpgen.auto_test.refine import make_refine
from dpgen.auto_test.reproduce import make_repro
from dpgen.auto_test.reproduce import post_repro


class Interstitial(Property):
    def __init__(self,
                 parameter):
        parameter['reproduce'] = parameter.get('reproduce', False)
        self.reprod = parameter['reproduce']
        if not self.reprod:
            if not ('init_from_suffix' in parameter and 'output_suffix' in parameter):
                default_supercell = [1, 1, 1]
                parameter['supercell'] = parameter.get('supercell', default_supercell)
                self.supercell = parameter['supercell']
                self.insert_ele = parameter['insert_ele']
            parameter['cal_type'] = parameter.get('cal_type', 'relaxation')
            self.cal_type = parameter['cal_type']
            default_cal_setting = {"relax_pos": True,
                                   "relax_shape": True,
                                   "relax_vol": True}
            if 'cal_setting' not in parameter:
                parameter['cal_setting'] = default_cal_setting
            else:
                if "relax_pos" not in parameter['cal_setting']:
                    parameter['cal_setting']['relax_pos'] = default_cal_setting['relax_pos']
                if "relax_shape" not in parameter['cal_setting']:
                    parameter['cal_setting']['relax_shape'] = default_cal_setting['relax_shape']
                if "relax_vol" not in parameter['cal_setting']:
                    parameter['cal_setting']['relax_vol'] = default_cal_setting['relax_vol']
            self.cal_setting = parameter['cal_setting']
        else:
            parameter['cal_type'] = 'static'
            self.cal_type = parameter['cal_type']
            default_cal_setting = {"relax_pos": False,
                                   "relax_shape": False,
                                   "relax_vol": False}
            if 'cal_setting' not in parameter:
                parameter['cal_setting'] = default_cal_setting
            else:
                if "relax_pos" not in parameter['cal_setting']:
                    parameter['cal_setting']['relax_pos'] = default_cal_setting['relax_pos']
                if "relax_shape" not in parameter['cal_setting']:
                    parameter['cal_setting']['relax_shape'] = default_cal_setting['relax_shape']
                if "relax_vol" not in parameter['cal_setting']:
                    parameter['cal_setting']['relax_vol'] = default_cal_setting['relax_vol']
            self.cal_setting = parameter['cal_setting']
            parameter['init_from_suffix'] = parameter.get('init_from_suffix', '00')
            self.init_from_suffix = parameter['init_from_suffix']
        self.parameter = parameter

    def make_confs(self,
                   path_to_work,
                   path_to_equi,
                   refine=False):
        path_to_work = os.path.abspath(path_to_work)
        path_to_equi = os.path.abspath(path_to_equi)

        if 'start_confs_path' in self.parameter and os.path.exists(self.parameter['start_confs_path']):
            init_path_list = glob.glob(os.path.join(self.parameter['start_confs_path'], '*'))
            struct_init_name_list = []
            for ii in init_path_list:
                struct_init_name_list.append(ii.split('/')[-1])
            struct_output_name = path_to_work.split('/')[-2]
            assert struct_output_name in struct_init_name_list
            path_to_equi = os.path.abspath(os.path.join(self.parameter['start_confs_path'],
                                                        struct_output_name, 'relaxation', 'relax_task'))

        task_list = []
        cwd = os.getcwd()

        if self.reprod:
            print('interstitial reproduce starts')
            if 'init_data_path' not in self.parameter:
                raise RuntimeError("please provide the initial data path to reproduce")
            init_data_path = os.path.abspath(self.parameter['init_data_path'])
            task_list = make_repro(init_data_path, self.init_from_suffix,
                                   path_to_work, self.parameter.get('reprod_last_frame', False))
            os.chdir(cwd)

        else:
            if refine:
                print('interstitial refine starts')
                task_list = make_refine(self.parameter['init_from_suffix'],
                                        self.parameter['output_suffix'],
                                        path_to_work)

                init_from_path = re.sub(self.parameter['output_suffix'][::-1],
                                        self.parameter['init_from_suffix'][::-1],
                                        path_to_work[::-1], count=1)[::-1]
                task_list_basename = list(map(os.path.basename, task_list))

                os.chdir(path_to_work)
                if os.path.isfile('element.out'):
                    os.remove('element.out')
                if os.path.islink('element.out'):
                    os.remove('element.out')
                os.symlink(os.path.relpath(os.path.join(init_from_path, 'element.out')), 'element.out')
                os.chdir(cwd)

                for ii in task_list_basename:
                    init_from_task = os.path.join(init_from_path, ii)
                    output_task = os.path.join(path_to_work, ii)
                    os.chdir(output_task)
                    if os.path.isfile('supercell.json'):
                        os.remove('supercell.json')
                    if os.path.islink('supercell.json'):
                        os.remove('supercell.json')
                    os.symlink(os.path.relpath(os.path.join(init_from_task, 'supercell.json')), 'supercell.json')
                os.chdir(cwd)

            else:
                equi_contcar = os.path.join(path_to_equi, 'CONTCAR')
                if not os.path.exists(equi_contcar):
                    raise RuntimeError("please do relaxation first")

                ss = Structure.from_file(equi_contcar)
                # gen defects
                dss = []
                insert_element_task = os.path.join(path_to_work, 'element.out')
                if os.path.isfile(insert_element_task):
                    os.remove(insert_element_task)

                for ii in self.insert_ele:
                    vds = InterstitialGenerator(ss, ii)
                    for jj in vds:
                        temp = jj.generate_defect_structure(self.supercell)
                        smallest_distance = list(set(temp.distance_matrix.ravel()))[1]
                        if 'conf_filters' in self.parameter and 'min_dist' in self.parameter['conf_filters']:
                            min_dist = self.parameter['conf_filters']['min_dist']
                            if smallest_distance >= min_dist:
                                dss.append(temp)
                                with open(insert_element_task, 'a+') as fout:
                                    print(ii, file=fout)
                        else:
                            dss.append(temp)
                            with open(insert_element_task, 'a+') as fout:
                                print(ii, file=fout)
                #            dss.append(jj.generate_defect_structure(self.supercell))

                print(
                    'gen interstitial with supercell ' + str(self.supercell) + ' with element ' + str(self.insert_ele))
                os.chdir(path_to_work)
                if os.path.isfile('POSCAR'):
                    os.remove('POSCAR')
                if os.path.islink('POSCAR'):
                    os.remove('POSCAR')
                os.symlink(os.path.relpath(equi_contcar), 'POSCAR')
                #           task_poscar = os.path.join(output, 'POSCAR')

                for ii in range(len(dss)):
                    output_task = os.path.join(path_to_work, 'task.%06d' % ii)
                    os.makedirs(output_task, exist_ok=True)
                    os.chdir(output_task)
                    for jj in ['INCAR', 'POTCAR', 'POSCAR', 'conf.lmp', 'in.lammps']:
                        if os.path.exists(jj):
                            os.remove(jj)
                    task_list.append(output_task)
                    dss[ii].to('POSCAR', 'POSCAR')
                    # np.savetxt('supercell.out', self.supercell, fmt='%d')
                    dumpfn(self.supercell, 'supercell.json')
                os.chdir(cwd)


                if 'bcc_self' in self.parameter and self.parameter['bcc_self']:
                    os.chdir(path_to_work)
                    with open('POSCAR', 'r') as fin:
                        fin.readline()
                        scale = float(fin.readline().split()[0])
                        latt_param = float(fin.readline().split()[0])
                        latt_param *= scale
                    
                    if not os.path.isfile('task.000000/POSCAR'):
                        raise RuntimeError("need task.000000 structure as reference")

                    with open('task.000000/POSCAR', 'r') as fin:
                        pos_line = fin.read().split('\n')

                    super_latt_param = float(pos_line[2].split()[0])

                    output_task1 = os.path.join(path_to_work, 'task.%06d' % (len(dss)))
                    os.makedirs(output_task1, exist_ok=True)
                    os.chdir(output_task1)
                    task_list.append(output_task1)
                    with open(insert_element_task, 'a+') as fout:
                        print(self.insert_ele[0], file=fout)
                    dumpfn(self.supercell, 'supercell.json')
                    pos_line[-2] = '%.6f' % float(latt_param/4/super_latt_param) + ' ' + '%.6f' % float(latt_param/2/super_latt_param) + ' 0.000000 ' + self.insert_ele[0]
                    with open('POSCAR', 'w+') as fout:
                        for ii in pos_line:
                                print(ii, file=fout)
                    print('gen bcc tetrahedral')
                    os.chdir(cwd)

                    output_task2 = os.path.join(path_to_work, 'task.%06d' % (len(dss)+1))
                    os.makedirs(output_task2, exist_ok=True)
                    os.chdir(output_task2)
                    task_list.append(output_task2)
                    with open(insert_element_task, 'a+') as fout:
                        print(self.insert_ele[0], file=fout)
                    dumpfn(self.supercell, 'supercell.json')
                    pos_line[-2] = '%.6f' % float(latt_param/2/super_latt_param) + ' ' + '%.6f' % float(latt_param/2/super_latt_param) + ' 0.000000 ' + self.insert_ele[0]
                    with open('POSCAR', 'w+') as fout:
                        for ii in pos_line:
                                print(ii, file=fout)
                    print('gen bcc octahedral')
                    os.chdir(cwd)

                    output_task3 = os.path.join(path_to_work, 'task.%06d' % (len(dss)+2))
                    os.makedirs(output_task3, exist_ok=True)
                    os.chdir(output_task3)
                    task_list.append(output_task3)
                    with open(insert_element_task, 'a+') as fout:
                        print(self.insert_ele[0], file=fout)
                    dumpfn(self.supercell, 'supercell.json')
                    pos_line[-2] = '%.6f' % float(latt_param/4/super_latt_param) + ' ' + '%.6f' % float(latt_param/4/super_latt_param) + ' ' + '%.6f' % float(latt_param/4/super_latt_param) + ' ' + self.insert_ele[0]
                    with open('POSCAR', 'w+') as fout:
                        for ii in pos_line:
                                print(ii, file=fout)
                    print('gen bcc crowdion')
                    os.chdir(cwd)

                    for idx, ii in enumerate(pos_line):
                        ss = ii.split()
                        if len(ss) > 3:
                            if abs(latt_param/2/super_latt_param - float(ss[0])) < 1e-5 and abs(latt_param/2/super_latt_param - float(ss[1])) < 1e-5 and abs(latt_param/2/super_latt_param - float(ss[2])) < 1e-5:
                                replace_label = idx

                    output_task4 = os.path.join(path_to_work, 'task.%06d' % (len(dss)+3))
                    os.makedirs(output_task4, exist_ok=True)
                    os.chdir(output_task4)
                    task_list.append(output_task4)
                    with open(insert_element_task, 'a+') as fout:
                        print(self.insert_ele[0], file=fout)
                    dumpfn(self.supercell, 'supercell.json')
                    pos_line[-2] = '%.6f' % float(latt_param/3/super_latt_param) + ' ' + '%.6f' % float(latt_param/3/super_latt_param) + ' ' + '%.6f' % float(latt_param/3/super_latt_param) + ' ' + self.insert_ele[0]
                    pos_line[replace_label] = '%.6f' % float(latt_param/3*2/super_latt_param) + ' ' + '%.6f' % float(latt_param/3*2/super_latt_param) + ' ' + '%.6f' % float(latt_param/3*2/super_latt_param) + ' ' + self.insert_ele[0]

                    with open('POSCAR', 'w+') as fout:
                        for ii in pos_line:
                                print(ii, file=fout)
                    print('gen bcc <111> dumbbell')
                    os.chdir(cwd)

                    output_task5 = os.path.join(path_to_work, 'task.%06d' % (len(dss)+4))
                    os.makedirs(output_task5, exist_ok=True)
                    os.chdir(output_task5)
                    task_list.append(output_task5)
                    with open(insert_element_task, 'a+') as fout:
                        print(self.insert_ele[0], file=fout)
                    dumpfn(self.supercell, 'supercell.json')
                    pos_line[-2] = '%.6f' % float((latt_param+2.1/2**0.5)/2/super_latt_param) + ' ' + '%.6f' % float((latt_param-2.1/2**0.5)/2/super_latt_param) + ' ' + '%.6f' % float(latt_param/2/super_latt_param) + ' ' + self.insert_ele[0]
                    pos_line[replace_label] = '%.6f' % float((latt_param-2.1/2**0.5)/2/super_latt_param) + ' ' + '%.6f' % float((latt_param+2.1/2**0.5)/2/super_latt_param) + ' ' + '%.6f' % float(latt_param/2/super_latt_param) + ' ' + self.insert_ele[0]

                    with open('POSCAR', 'w+') as fout:
                        for ii in pos_line:
                                print(ii, file=fout)
                    print('gen bcc <110> dumbbell')
                    os.chdir(cwd)

                    output_task6 = os.path.join(path_to_work, 'task.%06d' % (len(dss)+5))
                    os.makedirs(output_task6, exist_ok=True)
                    os.chdir(output_task6)
                    task_list.append(output_task6)
                    with open(insert_element_task, 'a+') as fout:
                        print(self.insert_ele[0], file=fout)
                    dumpfn(self.supercell, 'supercell.json')
                    pos_line[-2] = '%.6f' % float(latt_param/2/super_latt_param) + ' ' + '%.6f' % float(latt_param/2/super_latt_param) + ' ' + '%.6f' % float((latt_param-2.1)/2/super_latt_param) + ' ' + self.insert_ele[0]
                    pos_line[replace_label] = '%.6f' % float(latt_param/2/super_latt_param) + ' ' + '%.6f' % float(latt_param/2/super_latt_param) + ' ' + '%.6f' % float((latt_param+2.1)/2/super_latt_param) + ' ' + self.insert_ele[0]

                    with open('POSCAR', 'w+') as fout:
                        for ii in pos_line:
                                print(ii, file=fout)
                    print('gen bcc <100> dumbbell')
                    os.chdir(cwd)


        return task_list

    def post_process(self, task_list):
        if True:
            fin1 = open(os.path.join(task_list[0], '..', 'element.out'), 'r')
            for ii in task_list:
                conf = os.path.join(ii, 'conf.lmp')
                inter = os.path.join(ii, 'inter.json')
                insert_ele = fin1.readline().split()[0]
                if os.path.isfile(conf):
                    with open(conf, 'r') as fin2:
                        conf_line = fin2.read().split('\n')
                        insert_line = conf_line[-2]
                    type_map = loadfn(inter)['type_map']
                    type_map_list = lammps.element_list(type_map)
                    if int(insert_line.split()[1]) > len(type_map_list):
                        type_num = type_map[insert_ele] + 1
                        conf_line[2] = str(len(type_map_list)) + ' atom types'
                        conf_line[-2] = '%6.d' % int(insert_line.split()[0]) + '%7.d' % type_num + \
                                        '%16.10f' % float(insert_line.split()[2]) + \
                                        '%16.10f' % float(insert_line.split()[3]) + \
                                        '%16.10f' % float(insert_line.split()[4])
                        with open(conf, 'w+') as fout:
                            for jj in conf_line:
                                print(jj, file=fout)
            fin1.close()

    def task_type(self):
        return self.parameter['type']

    def task_param(self):
        return self.parameter

    def _compute_lower(self,
                       output_file,
                       all_tasks,
                       all_res):
        output_file = os.path.abspath(output_file)
        res_data = {}
        ptr_data = os.path.dirname(output_file) + '\n'

        if not self.reprod:
            with open(os.path.join(os.path.dirname(output_file), 'element.out'), 'r') as fin:
                fc = fin.read().split('\n')
            ptr_data += "Insert_ele-Struct: Inter_E(eV)  E(eV) equi_E(eV)\n"
            idid = -1
            for ii in all_tasks:
                idid += 1
                structure_dir = os.path.basename(ii)
                task_result = loadfn(all_res[idid])
                natoms = task_result['atom_numbs'][0]
                equi_path = os.path.abspath(os.path.join(os.path.dirname(output_file), '../relaxation/relax_task'))
                equi_result = loadfn(os.path.join(equi_path, 'result.json'))
                equi_epa = equi_result['energies'][-1] / equi_result['atom_numbs'][0]
                evac = task_result['energies'][-1] - equi_epa * natoms

                supercell_index = loadfn(os.path.join(ii, 'supercell.json'))
                # insert_ele = loadfn(os.path.join(ii, 'task.json'))['insert_ele'][0]
                insert_ele = fc[idid]
                ptr_data += "%s: %7.3f  %7.3f %7.3f \n" % (
                    insert_ele + '-' + str(supercell_index) + '-' + structure_dir, evac,
                    task_result['energies'][-1], equi_epa * natoms)
                res_data[insert_ele + '-' + str(supercell_index) + '-' + structure_dir] = [evac,
                                                                                           task_result['energies'][-1],
                                                                                           equi_epa * natoms]

        else:
            if 'init_data_path' not in self.parameter:
                raise RuntimeError("please provide the initial data path to reproduce")
            init_data_path = os.path.abspath(self.parameter['init_data_path'])
            res_data, ptr_data = post_repro(init_data_path, self.parameter['init_from_suffix'],
                                            all_tasks, ptr_data, self.parameter.get('reprod_last_frame', False))

        with open(output_file, 'w') as fp:
            json.dump(res_data, fp, indent=4)

        return res_data, ptr_data



import os,sys,json,glob,shutil
import unittest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__package__ = 'data'
from .context import setUpModule
from .context_bulk import *

class TestGenBulkABACUS(unittest.TestCase):
    def setUp(self):
        self.alloy=[] 
        with open (abacus_param_file, 'r') as fp :
            jdata = json.load (fp)
        out_dir = out_dir_name(jdata)
        self.out_dir= out_dir
        jdata['out_dir'] = out_dir
        self.elements=jdata["elements"]
        self.scale_numb=len(jdata["scale"])
        self.pert_numb=jdata["pert_numb"]
        self.root_dir= out_dir
        create_path(out_dir)
        stru_data = make_unit_cell_ABACUS(jdata)
        supercell_stru = make_super_cell_ABACUS(jdata, stru_data)            
        place_element_ABACUS(jdata, supercell_stru)
        make_abacus_relax(jdata, {"fp_resources":{}})
        make_scale_ABACUS(jdata)
        pert_scaled(jdata)

    def tearDown(self):
        shutil.rmtree(self.root_dir)

    def test(self):
        path=self.out_dir+"/00.place_ele"
        #struct0=Structure.from_file(os.path.join(path,"STRU"))
        alloys=glob.glob(os.path.join(path,"sys-*"))
        stru0 = get_abacus_STRU(os.path.join(alloys[0], "STRU"))
        self.assertEqual(len(alloys),stru0['coords'].shape[0]-1)
        for ii in alloys:
            elem_numb=[int(i) for i in ii.split('/')[-1].split('-')[1:]]
            struct=get_abacus_STRU(os.path.join(ii,"STRU"))
            self.assertEqual(struct["atom_names"],self.elements)
            self.assertEqual(struct["atom_numbs"], elem_numb)
        path=self.out_dir+"/01.scale_pert"
        alloys=glob.glob(os.path.join(path,"sys-*"))
        self.assertEqual(len(alloys), stru0['coords'].shape[0]-1)
        for ii in alloys:
            scales=glob.glob(os.path.join(ii,"scale-*"))
            self.assertEqual(len(scales),self.scale_numb)
            for scale in scales:
                perts=glob.glob(os.path.join(scale,"[0-9]*")) 
                self.assertEqual(len(perts),self.pert_numb+1)


if __name__ == '__main__':
    unittest.main()


import os,sys
import dpdata
import numpy as np
import unittest
import importlib

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__package__ = 'generator'
from .context import check_cluster
from .context import setUpModule


class Test_check_cluster(unittest.TestCase):
    def test (self) :
        conf_name='POSCAR_Au_cluster'
        fmt='POSCAR'
        ret = check_cluster(conf_name, fp_cluster_vacuum=15, fmt=fmt)
        #bad cluster
        self.assertTrue(ret)
        #good cluster 
        ret = check_cluster(conf_name, fp_cluster_vacuum=10, fmt=fmt)
        self.assertFalse(ret)

if __name__ == '__main__':
    unittest.main()
         





import os, sys, glob, shutil
import unittest
import json
import numpy as np
import importlib

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__package__ = 'generator'
dirname = os.path.join(os.path.abspath(os.path.dirname(__file__)), "gromacs")

from .context import make_model_devi
from .context import make_fp_gaussian

def _make_fake_graphs(train_path):
    if not os.path.exists(train_path):
        os.mkdir(train_path)
    for ii in range(4):
        with open(os.path.join(train_path, f"graph.{ii:03}.pb"), 'w+') as f:
            f.write("Fake Model")

class TestGromacsModelDeviEngine(unittest.TestCase):
    def setUp(self):
        self.dirname = dirname
        self.jdata = {
            "type_map": ["H", "C", "N", "O", "Cl"],
            "mass_map": [2, 12, 14, 16, 35],
            "sys_configs_prefix": self.dirname,
            "sys_configs": [["model_devi_case"]],
            "sys_format": "gromacs/gro",
            "model_devi_engine": "gromacs",
            "gromacs_settings": {
                "mdp_filename":      "md.mdp",
                "topol_filename":    "processed.top",
                "conf_filename":     "npt.gro",
                "index_filename":    "index.raw",
                "ref_filename":      "em.tpr",
                "model_devi_script": "model_devi.py",
                "traj_filename":     "deepmd_traj.gro"
            },
            "model_devi_dt":         0.001,
            "model_devi_f_trust_lo": 0.05,
            "model_devi_f_trust_hi": 0.10,
            "model_devi_e_trust_lo": 1e10,
            "model_devi_e_trust_hi": 1e10,
            "model_devi_clean_traj": False,
            "model_devi_skip":       0,
            "model_devi_nopbc":      True,
            "model_devi_jobs": [
                {
                    "ensemble": "nvt",
                    "nsteps":   5000,
                    "sys_idx":  [0],
                    "trj_freq": 10
                }
            ],
            "shuffle_poscar": False,
            "fp_style": "gaussian",
            "shuffle_poscar": False,
            "fp_task_max": 20,
            "fp_task_min": 1,
            "fp_pp_path": "./",
            "fp_pp_files": [],
            "fp_params": {
                "keywords": "force m062x/6-31g(d) nosymm",
                "nproc": 16,
                "multiplicity": "auto"
            }
        }
        self.iter_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "iter.000000")
        if not os.path.exists(self.iter_path):
            os.mkdir(self.iter_path)
        self.train_path = os.path.join(self.iter_path, "00.train")
        self.model_devi_path = os.path.join(self.iter_path, "01.model_devi")
        self.model_devi_task_path = os.path.join(self.model_devi_path, "task.000.000000")
        self.fp_path = os.path.join(self.iter_path, "02.fp")
        _make_fake_graphs(self.train_path)
    
    def _check_dir(self, wdir, post=True):
        for key in self.jdata['gromacs_settings'].keys():
            if key != "traj_filename":
                self.assertTrue(os.path.exists(os.path.join(wdir, self.jdata['gromacs_settings'][key])))
            else:
                if post:
                    self.assertTrue(os.path.exists(os.path.join(wdir, self.jdata['gromacs_settings'][key])))
    
    def _copy_outputs(self, path_1, path_2):
        shutil.copy(os.path.join(path_1, "deepmd_traj.gro"), os.path.join(path_2, "deepmd_traj.gro"))
        shutil.copy(os.path.join(path_1, "model_devi.out"), os.path.join(path_2, "model_devi.out"))
        shutil.copytree(os.path.join(path_1, "traj"), os.path.join(path_2, "traj"))

    def test_make_model_devi_gromacs(self):
        flag = make_model_devi(iter_index=0,
                               jdata=self.jdata,
                               mdata={"deepmd_version": "2.0"})
        self.assertTrue(flag)
        self.assertTrue(os.path.exists(self.model_devi_path))
        self.assertTrue(os.path.exists(self.model_devi_task_path))
        self._check_dir(self.model_devi_task_path, post=False)
        self._copy_outputs(os.path.join(self.dirname, "outputs"), self.model_devi_task_path)
        self._check_dir(self.model_devi_task_path, post=True)
    
    @unittest.skipIf(importlib.util.find_spec("openbabel") is None, "requires openbabel")
    def test_make_fp_gaussian(self):
        make_fp_gaussian(iter_index=0, jdata=self.jdata)
        candi = np.loadtxt(os.path.join(self.fp_path, "candidate.shuffled.000.out"), dtype=np.str)
        candi_ref = np.loadtxt(os.path.join(self.dirname, "outputs", "candidate.shuffled.000.out"), dtype=np.str)
        self.assertEqual(sorted([int(i) for i in candi[:,1]]), [0,10,20,30,50])
        
     
    def tearDown(self):
        # pass
        shutil.rmtree(self.iter_path)
if __name__ == '__main__':
    unittest.main()




import glob
import os
import warnings
from monty.serialization import dumpfn
from multiprocessing import Pool

import dpgen.auto_test.lib.crys as crys
import dpgen.auto_test.lib.util as util
from dpgen import dlog
from dpgen.auto_test.calculator import make_calculator
from dpgen.auto_test.mpdb import get_structure
from dpgen.dispatcher.Dispatcher import make_dispatcher
from distutils.version import LooseVersion
from dpgen.dispatcher.Dispatcher import make_submission
from dpgen.remote.decide_machine import convert_mdata
from dpgen.auto_test.lib.utils import create_path

lammps_task_type = ['deepmd', 'meam', 'eam_fs', 'eam_alloy']


def make_equi(confs,
              inter_param,
              relax_param):
    # find all POSCARs and their name like mp-xxx
    # ...
    dlog.debug('debug info make equi')
    if 'type_map' in inter_param:
        ele_list = [key for key in inter_param['type_map'].keys()]
    else:
        ele_list = [key for key in inter_param['potcars'].keys()]
    # ele_list = inter_param['type_map']
    dlog.debug("ele_list %s" % ':'.join(ele_list))
    conf_dirs = []
    for conf in confs:
        conf_dirs.extend(glob.glob(conf))
    conf_dirs.sort()

    # generate a list of task names like mp-xxx/relaxation/relax_task
    # ...
    cwd = os.getcwd()
    # generate poscar for single element crystal
    if len(ele_list) == 1 or 'single' in inter_param:
        if 'single' in inter_param:
            element_label = int(inter_param['single'])
        else:
            element_label = 0
        for ii in conf_dirs:
            os.chdir(ii)
            crys_type = ii.split('/')[-1]
            dlog.debug('crys_type: %s' % crys_type)
            dlog.debug('pwd: %s' % os.getcwd())
            if crys_type == 'std-fcc':
                if not os.path.exists('POSCAR'):
                    crys.fcc1(ele_list[element_label]).to('POSCAR', 'POSCAR')
            elif crys_type == 'std-hcp':
                if not os.path.exists('POSCAR'):
                    crys.hcp(ele_list[element_label]).to('POSCAR', 'POSCAR')
            elif crys_type == 'std-dhcp':
                if not os.path.exists('POSCAR'):
                    crys.dhcp(ele_list[element_label]).to('POSCAR', 'POSCAR')
            elif crys_type == 'std-bcc':
                if not os.path.exists('POSCAR'):
                    crys.bcc(ele_list[element_label]).to('POSCAR', 'POSCAR')
            elif crys_type == 'std-diamond':
                if not os.path.exists('POSCAR'):
                    crys.diamond(ele_list[element_label]).to('POSCAR', 'POSCAR')
            elif crys_type == 'std-sc':
                if not os.path.exists('POSCAR'):
                    crys.sc(ele_list[element_label]).to('POSCAR', 'POSCAR')

            os.chdir(cwd)
    task_dirs = []
    # make task directories like mp-xxx/relaxation/relax_task
    # if mp-xxx/exists then print a warning and exit.
    # ...
    for ii in conf_dirs:
        crys_type = ii.split('/')[-1]
        dlog.debug('crys_type: %s' % crys_type)

        if 'mp-' in crys_type and not os.path.exists(os.path.join(ii, 'POSCAR')):
            get_structure(crys_type).to('POSCAR', os.path.join(ii, 'POSCAR'))

        poscar = os.path.abspath(os.path.join(ii, 'POSCAR'))
        if not os.path.exists(poscar):
            raise FileNotFoundError('no configuration for autotest')
        if os.path.exists(os.path.join(ii, 'relaxation', 'jr.json')):
            os.remove(os.path.join(ii, 'relaxation', 'jr.json'))

        relax_dirs = os.path.abspath(os.path.join(ii, 'relaxation', 'relax_task'))    # to be consistent with property in make dispatcher
        create_path(relax_dirs)
        task_dirs.append(relax_dirs)
        os.chdir(relax_dirs)
        # copy POSCARs to mp-xxx/relaxation/relax_task
        # ...
        if os.path.isfile('POSCAR'):
            os.remove('POSCAR')
        os.symlink(os.path.relpath(poscar), 'POSCAR')
        os.chdir(cwd)
    task_dirs.sort()
    # generate task files
    relax_param['cal_type'] = 'relaxation'
    if 'cal_setting' not in relax_param:
        relax_param['cal_setting'] = {"relax_pos": True,
                                      "relax_shape": True,
                                      "relax_vol": True}
    elif "relax_pos" not in relax_param['cal_setting']:
        relax_param['cal_setting']['relax_pos'] = True
    elif "relax_shape" not in relax_param['cal_setting']:
        relax_param['cal_setting']['relax_shape'] = True
    elif "relax_vol" not in relax_param['cal_setting']:
        relax_param['cal_setting']['relax_vol'] = True

    for ii in task_dirs:
        poscar = os.path.join(ii, 'POSCAR')
        dlog.debug('task_dir %s' % ii)
        inter = make_calculator(inter_param, poscar)
        inter.make_potential_files(ii)
        inter.make_input_file(ii, 'relaxation', relax_param)



def run_equi(confs,
             inter_param,
             mdata):
    # find all POSCARs and their name like mp-xxx
    # ...
    conf_dirs = []
    for conf in confs:
        conf_dirs.extend(glob.glob(conf))
    conf_dirs.sort()

    processes = len(conf_dirs)

    # generate a list of task names like mp-xxx/relaxation/relax_task
    # ...
    work_path_list = []
    for ii in conf_dirs:
        work_path_list.append(os.path.join(ii, 'relaxation'))
    all_task = []
    for ii in work_path_list:
        all_task.append(os.path.join(ii, 'relax_task'))
    run_tasks = all_task
    inter_type = inter_param['type']
    # vasp
    if inter_type == "vasp":
        mdata = convert_mdata(mdata, ["fp"])
    elif inter_type in lammps_task_type:
        mdata = convert_mdata(mdata, ["model_devi"])
    else:
        raise RuntimeError("unknown task %s, something wrong" % inter_type)
    
    # dispatch the tasks
    # POSCAR here is useless
    virtual_calculator = make_calculator(inter_param, "POSCAR")
    forward_files = virtual_calculator.forward_files()
    forward_common_files = virtual_calculator.forward_common_files()
    backward_files = virtual_calculator.backward_files()
    #    backward_files += logs
    machine, resources, command, group_size = util.get_machine_info(mdata, inter_type)
    work_path = os.getcwd()
    print("%s --> Runing... " % (work_path))

    api_version = mdata.get('api_version', '0.9')
    if LooseVersion(api_version) < LooseVersion('1.0'):
        warnings.warn(f"the dpdispatcher will be updated to new version."
                      f"And the interface may be changed. Please check the documents for more details")
        disp = make_dispatcher(machine, resources, work_path, run_tasks, group_size)
        disp.run_jobs(resources,
                      command,
                      work_path,
                      run_tasks,
                      group_size,
                      forward_common_files,
                      forward_files,
                      backward_files,
                      outlog='outlog',
                      errlog='errlog')
    elif LooseVersion(api_version) >= LooseVersion('1.0'):
    
        submission = make_submission(
            mdata_machine=machine,
            mdata_resources=resources,
            commands=[command],
            work_path=work_path,
            run_tasks=run_tasks,
            group_size=group_size,
            forward_common_files=forward_common_files,
            forward_files=forward_files,
            backward_files=backward_files,
            outlog='outlog',
            errlog='errlog'
        )
        submission.run_submission()
    
    

def post_equi(confs, inter_param):
    # find all POSCARs and their name like mp-xxx
    # ...
    conf_dirs = []
    for conf in confs:
        conf_dirs.extend(glob.glob(conf))
    conf_dirs.sort()
    task_dirs = []
    for ii in conf_dirs:
        task_dirs.append(os.path.abspath(os.path.join(ii, 'relaxation', 'relax_task')))
    task_dirs.sort()

    # generate a list of task names like mp-xxx/relaxation
    # ...

    # dump the relaxation result.
    for ii in task_dirs:
        poscar = os.path.join(ii, 'POSCAR')
        inter = make_calculator(inter_param, poscar)
        res = inter.compute(ii)
        dumpfn(res, os.path.join(ii, 'result.json'), indent=4)



import numpy as np
import os

def _parse_calypso_input(var,input_path):
        try:
            f = open(input_path,'r')
        except:
            f = open(os.path.join(input_path,'input.dat'),'r')
        lines = f.readlines()
        f.close()

        for line in lines:
            if var in line:
                variable = line.split('=')[1].strip()
                return variable

def _parse_calypso_dis_mtx(numberofspecies,input_path):
        try:
            f = open(input_path,'r')
        except:
            f = open(os.path.join(input_path,'input.dat'),'r')
        while True:
            line = f.readline()
            if len(line) == 0:
                break
            if '@DistanceOfIon' in line: 
                dis = []
                for i in range(int(numberofspecies)):
                    line = f.readline()
                    dis.append(line.split())
                f.close()
                break
        dis = np.array(dis)
        dis = dis.reshape((1,int(numberofspecies)**2))
        return dis[0][np.argmin(dis)]



#!/usr/bin/env python3

import os, re, shutil, logging
import glob

iter_format = "%06d"
task_format = "%02d"
log_iter_head = "iter " + iter_format + " task " + task_format + ": "

def make_iter_name (iter_index) :
    return "iter." + (iter_format % iter_index)

def create_path (path) :
    path += '/'
    if os.path.isdir(path) : 
        dirname = os.path.dirname(path)        
        counter = 0
        while True :
            bk_dirname = dirname + ".bk%03d" % counter
            if not os.path.isdir(bk_dirname) : 
                shutil.move (dirname, bk_dirname) 
                break
            counter += 1
    os.makedirs (path)

def replace (file_name, pattern, subst) :
    file_handel = open (file_name, 'r')
    file_string = file_handel.read ()
    file_handel.close ()
    file_string = ( re.sub (pattern, subst, file_string) )
    file_handel = open (file_name, 'w')
    file_handel.write (file_string)
    file_handel.close ()

def copy_file_list (file_list, from_path, to_path) :
    for jj in file_list : 
        if os.path.isfile(os.path.join(from_path, jj)) :
            shutil.copy (os.path.join(from_path, jj), to_path)
        elif os.path.isdir(os.path.join(from_path, jj)) :
            shutil.copytree (os.path.join(from_path, jj), os.path.join(to_path, jj))

def cmd_append_log (cmd,
                    log_file) :
    ret = cmd
    ret = ret + " 1> " + log_file
    ret = ret + " 2> " + log_file
    return ret

def log_iter (task, ii, jj) :
    logging.info ((log_iter_head + "%s") % (ii, jj, task))

def repeat_to_length(string_to_expand, length):
    ret = ""
    for ii in range (length) : 
        ret += string_to_expand
    return ret

def log_task (message) :
    header = repeat_to_length (" ", len(log_iter_head % (0, 0)))
    logging.info (header + message)

def record_iter (record, ii, jj) :
    with open (record, "a") as frec :
        frec.write ("%d %d\n" % (ii, jj))

def symlink_user_forward_files(mdata, task_type, work_path, task_format = None):
    '''
    Symlink user-defined forward_common_files
    Current path should be work_path, such as 00.train
    
    Parameters
    ---------
    mdata : dict
        machine parameters
    task_type: str
        task_type, such as "train"
    work_path : str
        work_path, such as "iter.000001/00.train"
    Returns
    -------
    None
    '''
    user_forward_files = mdata.get(task_type + "_" + "user_forward_files", [])
    #Angus: In the future, we may unify the task format.
    if task_format is None:
        task_format = {"train" : "0*", "model_devi" : "task.*", "fp": "task.*"}
        #"init_relax" : "sys-*", "init_md" : "sys-*/scale*/00*"
    for file in user_forward_files:
        assert os.path.isfile(file)  ,\
            "user_forward_file %s of %s stage doesn't exist. " % (file, task_type)
        tasks = glob.glob(os.path.join(work_path, task_format[task_type]))
        for task in tasks:
            if os.path.isfile(os.path.join(task, os.path.basename(file))):
                os.remove(os.path.join(task, os.path.basename(file)))
            os.symlink(file, os.path.join(task, os.path.basename(file)))
    return
    


"""
calypso as model devi engine:
       1. gen_structures
       2. analysis
       3. model devi
"""

import copy
import dpdata
import math
import numpy as np
import os
import random
import re
import glob
import shutil
from ase.io.vasp import write_vasp
from ase.io.trajectory import Trajectory
from pathlib import Path
from distutils.version import LooseVersion
from dpgen import dlog
from dpgen.generator.lib.utils import make_iter_name
from dpgen.generator.lib.parse_calypso import _parse_calypso_input
from dpgen.dispatcher.Dispatcher import make_dispatcher, make_submission

train_name = '00.train'
model_devi_name = '01.model_devi'
fp_name = '02.fp'
calypso_run_opt_name = 'gen_stru_analy'
calypso_model_devi_name = 'model_devi_results'

def gen_structures(iter_index,jdata,mdata):

    # run calypso
    # vsc means generate elemental, binary and ternary at the same time
    vsc = jdata.get('vsc',False)  # take CALYPSO as confs generator

    model_devi_group_size = mdata['model_devi_group_size']
    model_devi_resources = mdata['model_devi_resources']
    api_version = mdata.get('api_version', '0.9')


    iter_name = make_iter_name(iter_index)
    work_path = os.path.join(iter_name, model_devi_name)
    assert(os.path.isdir(work_path))
    # 
    calypso_run_opt_path = os.path.join(work_path,calypso_run_opt_name)
    calypso_model_devi_path = os.path.join(work_path,calypso_model_devi_name)
    #

    calypso_path = mdata.get('model_devi_calypso_path')
    #calypso_input_path = jdata.get('calypso_input_path')
    popsize = int(_parse_calypso_input('PopSize',calypso_run_opt_path))
    maxstep = int(_parse_calypso_input('MaxStep',calypso_run_opt_path))
    
    all_models = glob.glob(os.path.join(calypso_run_opt_path, 'graph*pb'))
    model_names = [os.path.basename(ii) for ii in all_models]

    deepmdkit_python = mdata.get('model_devi_deepmdkit_python')
    command = "%s calypso_run_opt.py %s 1>> model_devi.log 2>> model_devi.log" % (deepmdkit_python,os.path.abspath(calypso_run_opt_path))
    command += "  ||  %s check_outcar.py %s " % (deepmdkit_python,os.path.abspath(calypso_run_opt_path))
    commands = [command]

    cwd = os.getcwd()
    os.chdir(calypso_run_opt_path)

    forward_files = ['POSCAR', 'calypso_run_opt.py','check_outcar.py','input.dat']
    backward_files = ['OUTCAR','CONTCAR','traj.traj','model_devi.log']

    run_calypso = calypso_path+'/calypso.x | tee log'
    if not vsc:
        Lpickup = _parse_calypso_input('PickUp','.')
        PickUpStep =  _parse_calypso_input('PickStep','.')
        if os.path.exists('tag_pickup_%s'%(str(PickUpStep))):
            dlog.info('caution! tag_pickup_%s exists!'%str(PickUpStep))
            Lpickup = 'F'
        if Lpickup == 'T':
            ftag = open('tag_pickup_%s'%(str(PickUpStep)),'w')
            ftag.close()
            os.remove('step')
            fstep = open('step','w')
            fstep.write('%12s'%str(PickUpStep))
            fstep.close()
        else:
            PickUpStep = 1
            try:
                os.mkdir('opt')
            except:
                pass
                #shutil.rmtree('opt')
                #os.mkdir('opt')


        for ii in range(int(PickUpStep)-1,maxstep+1):
            dlog.info('CALYPSO step %s'%ii)
            if ii == maxstep :  
                while True:
                    if len(glob.glob('OUTCAR_*')) == popsize:
                        break
                os.system('%s'%run_calypso)
                break
            # run calypso

            os.system('%s'%(run_calypso))
            
            for pop in range(ii*int(popsize),(ii+1)*int(popsize)):
                try:
                    os.mkdir('task.%03d'%pop)
                except:
                    shutil.rmtree('task.%03d'%pop)
                    os.mkdir('task.%03d'%pop)
                shutil.copyfile('calypso_run_opt.py',os.path.join('task.%03d'%pop,'calypso_run_opt.py'))
                shutil.copyfile('check_outcar.py',os.path.join('task.%03d'%pop,'check_outcar.py'))
                shutil.copyfile('POSCAR_%s'%str(pop-ii*int(popsize)+1),os.path.join('task.%03d'%(pop),'POSCAR'))
                shutil.copyfile('input.dat',os.path.join('task.%03d'%pop,'input.dat'))
            #for iii in range(1,popsize+1): 
            #    shutil.copyfile('POSCAR_%s'%str(iii),os.path.join('task.%03d'%(iii-1),'POSCAR'))

            all_task = glob.glob( "task.*")
            all_task.sort()

            run_tasks_ = all_task

            run_tasks = [os.path.basename(ii) for ii in run_tasks_]

            if LooseVersion(api_version) < LooseVersion('1.0'):
                warnings.warn(f"the dpdispatcher will be updated to new version."
                    f"And the interface may be changed. Please check the documents for more details")
                dispatcher=make_dispatcher(mdata['model_devi_machine'],mdata['model_devi_resources'],'./', run_tasks, model_devi_group_size)
                dispatcher.run_jobs(mdata['model_devi_resources'],
                                commands,
                                './',
                                run_tasks,
                                model_devi_group_size,
                                model_names,
                                forward_files,
                                backward_files,
                                outlog = 'model_devi.log',
                                errlog = 'model_devi.log')
            elif LooseVersion(api_version) >= LooseVersion('1.0'):
                os.chdir(cwd)
                submission = make_submission(
                    mdata['model_devi_machine'],
                    mdata['model_devi_resources'],
                    commands=commands,
                    work_path=calypso_run_opt_path,
                    run_tasks=run_tasks,
                    group_size=model_devi_group_size,
                    forward_common_files=model_names,
                    forward_files=forward_files,
                    backward_files=backward_files,
                    outlog = 'model_devi.log',
                    errlog = 'model_devi.log')
                submission.run_submission()
                os.chdir(calypso_run_opt_path)
           
            sstep = os.path.join('opt',str(ii))
            os.mkdir(sstep)
            if not os.path.exists('traj'):
                os.mkdir('traj')

            for jjj in range(ii*int(popsize),(ii+1)*int(popsize)):
                # to opt directory
                shutil.copyfile('POSCAR_%s'%str(jjj+1-ii*int(popsize)),os.path.join(sstep,'POSCAR_%s'%str(jjj+1-ii*int(popsize))),)
                shutil.copyfile(os.path.join('task.%03d'%(jjj),'OUTCAR'),os.path.join(sstep,'OUTCAR_%s'%str(jjj+1-ii*int(popsize))),)
                shutil.copyfile(os.path.join('task.%03d'%(jjj),'CONTCAR'),os.path.join(sstep,'CONTCAR_%s'%str(jjj+1-ii*int(popsize))),)
                # to run calypso directory
                shutil.copyfile(os.path.join('task.%03d'%(jjj),'OUTCAR'),'OUTCAR_%s'%str(jjj+1-ii*int(popsize)),)
                shutil.copyfile(os.path.join('task.%03d'%(jjj),'CONTCAR'),'CONTCAR_%s'%str(jjj+1-ii*int(popsize)),)
                # to traj
                shutil.copyfile(os.path.join('task.%03d'%(jjj),'traj.traj'),os.path.join('traj','%s.traj'%str(jjj+1)),)

            if LooseVersion(api_version) < LooseVersion('1.0'):
                os.rename('jr.json','jr_%s.json'%(str(ii)))

            tlist = glob.glob('task.*')
            for t in tlist:
                shutil.rmtree(t)

    else:
        # --------------------------------------------------------------
        # TODO(zhenyu) make this code work for other situation 
        component = ['Mg','Al','Cu','MgAl','MgCu','AlCu','MgAlCu']
        
        for idx,com in enumerate(component):
            pwd = os.getcwd()
            os.mkdir(str(idx))
            shutil.copyfile(os.path.join(cwd,'calypso_input','input.dat.%s'%com),os.path.join(str(idx),'input.dat'))
            os.chdir(str(idx))
            os.system(run_calypso)
            os.chdir(pwd)
        
        name_list = Path('.').glob('*/POSCAR_*')
        for idx,name in enumerate(name_list):
            shutil.copyfile(name,'POSCAR_%s'%(idx+1))
            try:
                os.mkdir('task.%04d'%(idx+1))
            except:
                shutil.rmtree('task.%04d'%(idx+1))
                os.mkdir('task.%04d'%(idx+1))
            shutil.copyfile('calypso_run_opt.py',os.path.join('task.%04d'%(idx+1),'calypso_run_opt.py'))
            shutil.copyfile('check_outcar.py',os.path.join('task.%04d'%(idx+1),'check_outcar.py'))
            shutil.copyfile('POSCAR_%s'%str(idx+1),os.path.join('task.%04d'%(idx+1),'POSCAR'))
            shutil.copyfile('input.dat',os.path.join('task.%04d'%(idx+1),'input.dat'))

        all_task = glob.glob( "task.*")
        all_task.sort()

        run_tasks_ = all_task

        run_tasks = [os.path.basename(ii) for ii in run_tasks_]

        if LooseVersion(api_version) < LooseVersion('1.0'):
            warnings.warn(f"the dpdispatcher will be updated to new version."
                f"And the interface may be changed. Please check the documents for more details")
            dispatcher=make_dispatcher(mdata['model_devi_machine'],mdata['model_devi_resources'],'./', run_tasks, model_devi_group_size)
            dispatcher.run_jobs(mdata['model_devi_resources'],
                            commands,
                            './',
                            run_tasks,
                            model_devi_group_size,
                            model_names,
                            forward_files,
                            backward_files,
                            outlog = 'model_devi.log',
                            errlog = 'model_devi.log')
        elif LooseVersion(api_version) >= LooseVersion('1.0'):
            os.chdir(cwd)
            submission = make_submission(
                mdata['model_devi_machine'],
                mdata['model_devi_resources'],
                commands=commands,
                work_path=calypso_run_opt_path,
                run_tasks=run_tasks,
                group_size=model_devi_group_size,
                forward_common_files=model_names,
                forward_files=forward_files,
                backward_files=backward_files,
                outlog = 'model_devi.log',
                errlog = 'model_devi.log')
            submission.run_submission()
            os.chdir(calypso_run_opt_path)
        
        os.mkdir('opt')
        if not os.path.exists('traj'):
            os.mkdir('traj')
        for jjj in range(len(all_task)):
            # to opt directory
            shutil.copyfile('POSCAR_%s'%str(jjj+1),os.path.join('opt','POSCAR_%s'%str(jjj+1)),)
            shutil.copyfile(os.path.join('task.%04d'%(jjj+1),'OUTCAR'),os.path.join('opt','OUTCAR_%s'%str(jjj+1)),)
            shutil.copyfile(os.path.join('task.%04d'%(jjj+1),'CONTCAR'),os.path.join('opt','CONTCAR_%s'%str(jjj+1)),)
            # to run calypso directory
            shutil.copyfile(os.path.join('task.%04d'%(jjj+1),'OUTCAR'),'OUTCAR_%s'%str(jjj+1),)
            shutil.copyfile(os.path.join('task.%04d'%(jjj+1),'CONTCAR'),'CONTCAR_%s'%str(jjj+1),)
            # to traj
            shutil.copyfile(os.path.join('task.%04d'%(jjj+1),'traj.traj'),os.path.join('traj','%s.traj'%str(jjj+1)),)

        tlist = glob.glob('task.*')
        for t in tlist:
            shutil.rmtree(t)
        # --------------------------------------------------------------

    os.chdir(cwd)
    os.chdir(work_path)
    f = open('record.calypso','a+')
    f.write('2\n')
    f.close()
    os.chdir(cwd)

def analysis(iter_index,jdata,calypso_run_opt_path,calypso_model_devi_path):

    # Analysis


    #dlog.info('$$$$$$$$$$$$$$$ Analysis Started  $$$$$$$$$$$$$$$$$$')

    ms = dpdata.MultiSystems()

    cwd = os.getcwd()
    iter_name = make_iter_name(iter_index)
    work_path = os.path.join(iter_name, model_devi_name)
    result_path = os.path.join(calypso_run_opt_path,'results')

    # trajs to be model devi
    traj_path = os.path.join(calypso_run_opt_path,'traj')
    traj_list = glob.glob(traj_path+'/*.traj')

    #dlog.info('len(traj_list) %s'%str(len(traj_list)))

    # read confs from traj
    record_traj_num = 0
    for traj_name in traj_list:
        traj_num = os.path.basename(traj_name).split('.')[0]
        trajs_origin = Trajectory(traj_name)

        record_traj_num += len(trajs_origin)
        if len(trajs_origin) >= 20 :
           trajs = [trajs_origin[iii] for iii in [4,9,-10,-5,-1]]
        elif 5<=len(trajs_origin)<20:
           trajs = [trajs_origin[random.randint(1,len(trajs_origin)-1)] for iii in range(4)]
           trajs.append(trajs[-1])
        elif 3<= len(trajs_origin) <5:
           trajs = [trajs_origin[round((len(trajs_origin)-1)/2)] ]
           trajs.append(trajs[-1])
        elif len(trajs_origin) == 2:
           trajs = [trajs_origin[0],trajs_origin[-1] ]
        elif len(trajs_origin) == 1:
           trajs = [trajs_origin[0] ]
        else:
           pass
           
        for idx, traj in enumerate(trajs):
            write_vasp(os.path.join(
                traj_path,'%03d.%03d.poscar' % (
                    int(traj_num), int(idx)
                    )
                ),
                traj)
           
    traj_pos_list = glob.glob(traj_path+'/*.poscar')

    #dlog.info('traj_num %s'%str(len(traj_pos_list)))
    #dlog.info('total_traj_num %s'%str(record_traj_num))

    for npos in traj_pos_list:
        try:
            ms.append(dpdata.System(npos, type_map = jdata['type_map']))
        except Exception as e:
            dlog.info(npos,'failed : ',e)

    if len(ms) == 0:
        print('too little confs, ')
        return

    if os.path.exists(os.path.join(result_path,'deepmd')):
        shutil.rmtree(os.path.join(result_path,'deepmd'))
    ms.to_deepmd_raw(os.path.join(result_path,'deepmd'))
    ms.to_deepmd_npy(os.path.join(result_path,'deepmd'))
    

    split_lists = glob.glob(os.path.join(result_path,'deepmd','*'))
    for i,split_list in enumerate(split_lists):
        #ss = dpdata.System(split_list,fmt='deepmd')
        #for j in range(ss.get_nframes()):
        #    ss.to('vasp/poscar',os.path.join(split_list,'%03d.%03d.poscar'%(i,j)),frame_idx=j)
        strus_path = os.path.join(calypso_model_devi_path,'%03d.structures'%i)
        if not os.path.exists(strus_path):
            shutil.copytree(split_list,strus_path)
        else:
            shutil.rmtree(strus_path)
            shutil.copytree(split_list,strus_path)

    os.chdir(cwd)
    os.chdir(work_path)
    f = open('record.calypso','a+')
    f.write('3\n')
    f.close()
    os.chdir(cwd)


def run_calypso_model_devi (iter_index,
                    jdata,
                    mdata) :

    dlog.info('start running CALYPSO')


    iter_name = make_iter_name(iter_index)
    work_path = os.path.join(iter_name, model_devi_name)
    assert(os.path.isdir(work_path))

    calypso_run_opt_path = os.path.join(work_path,calypso_run_opt_name)
    calypso_model_devi_path = os.path.join(work_path,calypso_model_devi_name)

    cwd = os.getcwd()

    record_calypso_path = os.path.join(work_path,'record.calypso')
    while True:
        if not os.path.exists(record_calypso_path):
            f = open(record_calypso_path,'w')
            f.write('1\n')
            lines = '1'
            f.close()
        else:
            f = open(record_calypso_path,'r')
            lines = f.readlines()
            f.close()

        if lines[-1].strip().strip('\n') == '1':
            # Gen Structures
            gen_structures(iter_index,jdata,mdata)

        elif lines[-1].strip().strip('\n') == '2':
            # Analysis & to deepmd/raw
            analysis(iter_index,jdata,calypso_run_opt_path,calypso_model_devi_path)

        elif lines[-1].strip().strip('\n') == '3':
            # Model Devi
            _calypso_run_opt_path = os.path.abspath(calypso_run_opt_path)
            all_models = glob.glob(os.path.join(_calypso_run_opt_path, 'graph*pb'))
            cwd = os.getcwd()
            os.chdir(calypso_model_devi_path)
            args = ' '.join(['calypso_run_model_devi.py', '--all_models',' '.join(all_models),'--type_map',' '.join(jdata.get('type_map'))])
            deepmdkit_python = mdata.get('model_devi_deepmdkit_python')
            os.system(f'{deepmdkit_python} {args} ')
            #Modd(iter_index,calypso_model_devi_path,all_models,jdata)
            os.chdir(cwd)

        elif lines[-1].strip().strip('\n') == '4':
            #dlog.info('Model Devi is done.')
            break



import glob
import json
import os
import re

from monty.serialization import loadfn, dumpfn
from pymatgen.analysis.defects.generators import VacancyGenerator
from pymatgen.core.structure import Structure

from dpgen import dlog
from dpgen.auto_test.Property import Property
from dpgen.auto_test.refine import make_refine
from dpgen.auto_test.reproduce import make_repro
from dpgen.auto_test.reproduce import post_repro


class Vacancy(Property):
    def __init__(self,
                 parameter):
        parameter['reproduce'] = parameter.get('reproduce', False)
        self.reprod = parameter['reproduce']
        if not self.reprod:
            if not ('init_from_suffix' in parameter and 'output_suffix' in parameter):
                default_supercell = [1, 1, 1]
                parameter['supercell'] = parameter.get('supercell', default_supercell)
                self.supercell = parameter['supercell']
            parameter['cal_type'] = parameter.get('cal_type', 'relaxation')
            self.cal_type = parameter['cal_type']
            default_cal_setting = {"relax_pos": True,
                                   "relax_shape": True,
                                   "relax_vol": True}
            if 'cal_setting' not in parameter:
                parameter['cal_setting'] = default_cal_setting
            else:
                if "relax_pos" not in parameter['cal_setting']:
                    parameter['cal_setting']['relax_pos'] = default_cal_setting['relax_pos']
                if "relax_shape" not in parameter['cal_setting']:
                    parameter['cal_setting']['relax_shape'] = default_cal_setting['relax_shape']
                if "relax_vol" not in parameter['cal_setting']:
                    parameter['cal_setting']['relax_vol'] = default_cal_setting['relax_vol']
            self.cal_setting = parameter['cal_setting']
        else:
            parameter['cal_type'] = 'static'
            self.cal_type = parameter['cal_type']
            default_cal_setting = {"relax_pos": False,
                                   "relax_shape": False,
                                   "relax_vol": False}
            if 'cal_setting' not in parameter:
                parameter['cal_setting'] = default_cal_setting
            else:
                if "relax_pos" not in parameter['cal_setting']:
                    parameter['cal_setting']['relax_pos'] = default_cal_setting['relax_pos']
                if "relax_shape" not in parameter['cal_setting']:
                    parameter['cal_setting']['relax_shape'] = default_cal_setting['relax_shape']
                if "relax_vol" not in parameter['cal_setting']:
                    parameter['cal_setting']['relax_vol'] = default_cal_setting['relax_vol']
            self.cal_setting = parameter['cal_setting']
            parameter['init_from_suffix'] = parameter.get('init_from_suffix', '00')
            self.init_from_suffix = parameter['init_from_suffix']
        self.parameter = parameter

    def make_confs(self,
                   path_to_work,
                   path_to_equi,
                   refine=False):
        path_to_work = os.path.abspath(path_to_work)
        if os.path.exists(path_to_work):
            dlog.warning('%s already exists' % path_to_work)
        else:
            os.makedirs(path_to_work)
        path_to_equi = os.path.abspath(path_to_equi)

        if 'start_confs_path' in self.parameter and os.path.exists(self.parameter['start_confs_path']):
            init_path_list = glob.glob(os.path.join(self.parameter['start_confs_path'], '*'))
            struct_init_name_list = []
            for ii in init_path_list:
                struct_init_name_list.append(ii.split('/')[-1])
            struct_output_name = path_to_work.split('/')[-2]
            assert struct_output_name in struct_init_name_list
            path_to_equi = os.path.abspath(os.path.join(self.parameter['start_confs_path'],
                                                        struct_output_name, 'relaxation', 'relax_task'))

        task_list = []
        cwd = os.getcwd()

        if self.reprod:
            print('vacancy reproduce starts')
            if 'init_data_path' not in self.parameter:
                raise RuntimeError("please provide the initial data path to reproduce")
            init_data_path = os.path.abspath(self.parameter['init_data_path'])
            task_list = make_repro(init_data_path, self.init_from_suffix,
                                   path_to_work, self.parameter.get('reprod_last_frame', False))
            os.chdir(cwd)

        else:
            if refine:
                print('vacancy refine starts')
                task_list = make_refine(self.parameter['init_from_suffix'],
                                        self.parameter['output_suffix'],
                                        path_to_work)

                init_from_path = re.sub(self.parameter['output_suffix'][::-1],
                                        self.parameter['init_from_suffix'][::-1],
                                        path_to_work[::-1], count=1)[::-1]
                task_list_basename = list(map(os.path.basename, task_list))

                for ii in task_list_basename:
                    init_from_task = os.path.join(init_from_path, ii)
                    output_task = os.path.join(path_to_work, ii)
                    os.chdir(output_task)
                    if os.path.isfile('supercell.json'):
                        os.remove('supercell.json')
                    if os.path.islink('supercell.json'):
                        os.remove('supercell.json')
                    os.symlink(os.path.relpath(os.path.join(init_from_task, 'supercell.json')), 'supercell.json')
                os.chdir(cwd)
            else:
                equi_contcar = os.path.join(path_to_equi, 'CONTCAR')
                if not os.path.exists(equi_contcar):
                    raise RuntimeError("please do relaxation first")

                ss = Structure.from_file(equi_contcar)
                vds = VacancyGenerator(ss)
                dss = []
                for jj in vds:
                    dss.append(jj.generate_defect_structure(self.supercell))

                print('gen vacancy with supercell ' + str(self.supercell))
                os.chdir(path_to_work)
                if os.path.isfile('POSCAR'):
                    os.remove('POSCAR')
                if os.path.islink('POSCAR'):
                    os.remove('POSCAR')
                os.symlink(os.path.relpath(equi_contcar), 'POSCAR')
                #           task_poscar = os.path.join(output, 'POSCAR')

                for ii in range(len(dss)):
                    output_task = os.path.join(path_to_work, 'task.%06d' % ii)
                    os.makedirs(output_task, exist_ok=True)
                    os.chdir(output_task)
                    for jj in ['INCAR', 'POTCAR', 'POSCAR', 'conf.lmp', 'in.lammps']:
                        if os.path.exists(jj):
                            os.remove(jj)
                    task_list.append(output_task)
                    dss[ii].to('POSCAR', 'POSCAR')
                    # np.savetxt('supercell.out', self.supercell, fmt='%d')
                    dumpfn(self.supercell, 'supercell.json')
                os.chdir(cwd)
        return task_list

    def post_process(self, task_list):
        pass

    def task_type(self):
        return self.parameter['type']

    def task_param(self):
        return self.parameter

    def _compute_lower(self,
                       output_file,
                       all_tasks,
                       all_res):
        output_file = os.path.abspath(output_file)
        res_data = {}
        ptr_data = os.path.dirname(output_file) + '\n'

        if not self.reprod:
            ptr_data += "Structure: \tVac_E(eV)  E(eV) equi_E(eV)\n"
            idid = -1
            for ii in all_tasks:
                idid += 1
                structure_dir = os.path.basename(ii)
                task_result = loadfn(all_res[idid])
                natoms = task_result['atom_numbs'][0]
                equi_path = os.path.abspath(os.path.join(os.path.dirname(output_file), '../relaxation/relax_task'))
                equi_result = loadfn(os.path.join(equi_path, 'result.json'))
                equi_epa = equi_result['energies'][-1] / equi_result['atom_numbs'][0]
                evac = task_result['energies'][-1] - equi_epa * natoms

                supercell_index = loadfn(os.path.join(ii, 'supercell.json'))
                ptr_data += "%s: %7.3f  %7.3f %7.3f \n" % (str(supercell_index) + '-' + structure_dir,
                                                           evac, task_result['energies'][-1], equi_epa * natoms)
                res_data[str(supercell_index) + '-' + structure_dir] = [evac, task_result['energies'][-1],
                                                                        equi_epa * natoms]

        else:
            if 'init_data_path' not in self.parameter:
                raise RuntimeError("please provide the initial data path to reproduce")
            init_data_path = os.path.abspath(self.parameter['init_data_path'])
            res_data, ptr_data = post_repro(init_data_path, self.parameter['init_from_suffix'],
                                            all_tasks, ptr_data, self.parameter.get('reprod_last_frame', False))

        with open(output_file, 'w') as fp:
            json.dump(res_data, fp, indent=4)

        return res_data, ptr_data



#!/usr/bin/env python
# coding: utf-8

from dpgen.dispatcher.SSHContext import SSHSession
from dpgen.dispatcher.SSHContext import SSHContext
from dpgen.dispatcher.Slurm import Slurm
from dpgen.dispatcher.LSF import LSF
from dpgen import dlog
import os
import json
import numpy as np
from distutils.version import LooseVersion


def convert_mdata(mdata, task_types=["train", "model_devi", "fp"]):
    '''
    Convert mdata for DP-GEN main process.
    New convension is like mdata["fp"]["machine"],
    DP-GEN needs mdata["fp_machine"]

    Notice that we deprecate the function which can automatically select one most avalaible machine,
    since this function was only used by Angus, and only supports for Slurm.
    In the future this can be implemented.

    Parameters
    ----------
    mdata : dict
        Machine parameters to be converted.
    task_types : list of string
        Type of tasks, default is ["train", "model_devi", "fp"]

    Returns
    -------
    dict
        mdata converted
    '''
    for task_type in task_types:
        if task_type in mdata:
            if isinstance(mdata[task_type], dict):
                task_data = mdata[task_type]
            elif isinstance(mdata[task_type], (list, tuple)):
                task_data = mdata[task_type][0]
            else:
                raise TypeError("mdata/%s should be dict or list!" % task_type)
            for key, item in task_data.items():
                if "comments" not in key:
                    mdata[task_type + "_" + key] = item
            group_size = task_data["resources"].get("group_size", 1)
            if group_size == 1:
                group_size = task_data.get("group_size", 1)
            mdata[task_type + "_" + "group_size"] = group_size
    return mdata



# def decide_train_machine(mdata):
# 	if LooseVersion(mdata.get('api_version', '0.9')) >= LooseVersion('1.0'):
# 		mdata['train_group_size'] = mdata['train'][0]['resources']['group_size']
# 	if 'train' in mdata:
# 		continue_flag = False
# 		if 'record.machine' in os.listdir():
# 			try:
# 				with open('record.machine', 'r') as _infile:
# 					profile = json.load(_infile)
# 					if profile['purpose'] == 'train':
# 						mdata['train_machine'] = profile['machine']
# 						mdata['train_resources'] = profile['resources']
#
# 						if 'python_path' in profile:
# 							mdata['python_path'] = profile['python_path']
# 						if "group_size" in profile:
# 							mdata["train_group_size"] = profile["group_size"]
# 						if 'deepmd_version' in profile:
# 							mdata["deepmd_version"] = profile['deepmd_version']
# 						if 'command' in profile:
# 							mdata['train_command'] = profile["command"]
# 						continue_flag = True
# 			except Exception:
# 				pass
# 		if ("hostname" not in mdata["train"][0]["machine"]) or (len(mdata["train"]) == 1):
# 			mdata["train_machine"] = mdata["train"][0]["machine"]
# 			mdata["train_resources"] = mdata["train"][0]["resources"]
#
# 			if 'python_path' in mdata["train"][0]:
# 				mdata["python_path"] = mdata["train"][0]["python_path"]
# 			if "group_size" in mdata["train"][0]:
# 				mdata["train_group_size"] = mdata["train"][0]["group_size"]
# 			if 'deepmd_version' in mdata["train"][0]:
# 				mdata["deepmd_version"] = mdata["train"][0]["deepmd_version"]
# 			if 'command' in mdata["train"][0]:
# 				mdata["train_command"] = mdata["train"][0]["command"]
# 			continue_flag = True
#
# 		pd_flag = False
# 		pd_count_list =[]
# 		# pd for pending job in slurm
# 		# if we need to launch new machine_idxines
# 		if not continue_flag:
#
# 			#assert isinstance(mdata['train']['machine'], list)
# 			#assert isinstance(mdata['train']['resources'], list)
# 			#assert len(mdata['train']['machine']) == len(mdata['train']['resources'])
# 			# mdata['train'] is  a list
# 			for machine_idx in range(len(mdata['train'])):
# 				temp_machine = mdata['train'][machine_idx]['machine']
# 				temp_resources = mdata['train'][machine_idx]['resources']
# 				temp_ssh_sess = SSHSession(temp_machine)
# 				cwd = os.getcwd()
# 				temp_context = SSHContext(cwd, temp_ssh_sess)
# 				if temp_machine['machine_type'] == 'lsf':
# 					temp_batch = LSF(temp_context)
# 				else:
# 					temp_batch = Slurm(temp_context)
# 				# For other type of machines, please add them using 'elif'.
# 				# Here slurm is selected as the final choice in convinience.
# 				command = temp_batch._make_squeue(temp_machine, temp_resources)
# 				ret, stdin, stdout, stderr = temp_batch.context.block_call(command)
# 				pd_response = stdout.read().decode('utf-8').split("\n")
# 				pd_count = len(pd_response)
# 				temp_context.clean()
# 				## If there is no need to waiting for allocation
# 				if pd_count ==1:
# 					mdata['train_machine'] = temp_machine
# 					mdata['train_resources'] = temp_resources
#
# 					if 'python_path' in mdata['train'][machine_idx]:
# 						mdata['python_path'] = mdata['train'][machine_idx]['python_path']
# 					if 'group_size' in mdata['train'][machine_idx]:
# 						mdata['train_group_size'] = mdata['train'][machine_idx]['group_size']
# 					if 'deepmd_version' in mdata['train'][machine_idx]:
# 						mdata['deepmd_version'] = mdata['train'][machine_idx]['deepmd_version']
# 					if 'command' in mdata['train'][machine_idx]:
# 						mdata['train_command'] = mdata['train'][machine_idx]['command']
#
# 					## No need to wait
# 					pd_flag = True
# 					break
# 				else:
# 					pd_count_list.append(pd_count)
# 			if not pd_flag:
# 				## All machines need waiting, then compare waiting jobs
# 				## Select a machine which has fewest waiting jobs
# 				min_machine_idx = np.argsort(pd_count_list)[0]
# 				mdata['train_machine'] = mdata['train'][min_machine_idx]['machine']
# 				mdata['train_resources'] = mdata['train'][min_machine_idx]['resources']
#
# 				if 'python_path' in mdata['train'][min_machine_idx]:
# 					mdata['python_path'] = mdata['train'][min_machine_idx]['python_path']
# 				if "group_size" in mdata['train'][min_machine_idx]:
# 					mdata["train_group_size"] = mdata['train'][min_machine_idx]["group_size"]
# 				if 'deepmd_version' in mdata['train'][min_machine_idx]:
# 					mdata['deepmd_version'] = mdata['train'][min_machine_idx]["deepmd_version"]
# 				if 'command' in mdata['train'][min_machine_idx]:
# 					mdata['train_command'] = mdata['train'][min_machine_idx]['command']
#
# 			## Record which machine is selected
# 			with open("record.machine","w") as _outfile:
# 				profile = {}
# 				profile['purpose'] = 'train'
# 				profile['machine'] = mdata['train_machine']
# 				profile['resources'] = mdata['train_resources']
#
# 				if 'python_path' in mdata:
# 					profile['python_path'] = mdata['python_path']
# 				if "train_group_size" in mdata:
# 					profile["group_size"] = mdata["train_group_size"]
# 				if 'deepmd_version' in mdata:
# 					profile['deepmd_version'] = mdata['deepmd_version']
# 				if 'train_command' in mdata:
# 					profile['command'] = mdata['train_command']
#
# 				json.dump(profile, _outfile, indent = 4)
# 	return mdata
#
# def decide_model_devi_machine(mdata):
# 	if LooseVersion(mdata.get('api_version', '0.9')) >= LooseVersion('1.0'):
# 		mdata['model_devi_group_size'] = mdata['model_devi'][0]['resources']['group_size']
# 	if 'model_devi' in mdata:
# 		continue_flag = False
# 		if 'record.machine' in os.listdir():
# 			try:
# 				with open('record.machine', 'r') as _infile:
# 					profile = json.load(_infile)
# 					if profile['purpose'] == 'model_devi':
# 						mdata['model_devi_machine'] = profile['machine']
# 						mdata['model_devi_resources'] = profile['resources']
# 						mdata['model_devi_command'] = profile['command']
# 						mdata['model_devi_group_size'] = profile['group_size']
# 						continue_flag = True
# 			except Exception:
# 				pass
# 		if ("hostname" not in mdata["model_devi"][0]["machine"]) or (len(mdata["model_devi"]) == 1):
# 			mdata["model_devi_machine"] = mdata["model_devi"][0]["machine"]
# 			mdata["model_devi_resources"] = mdata["model_devi"][0]["resources"]
# 			mdata["model_devi_command"] = mdata["model_devi"][0]["command"]
# 			#if "group_size" in mdata["train"][0]:
# 			mdata["model_devi_group_size"] = mdata["model_devi"][0].get("group_size", 1)
# 			continue_flag = True
#
# 		pd_count_list =[]
# 		pd_flag = False
# 		if not continue_flag:
#
# 			#assert isinstance(mdata['model_devi']['machine'], list)
# 			#ssert isinstance(mdata['model_devi']['resources'], list)
# 			#assert len(mdata['model_devi']['machine']) == len(mdata['model_devi']['resources'])
#
# 			for machine_idx in range(len(mdata['model_devi'])):
# 				temp_machine = mdata['model_devi'][machine_idx]['machine']
# 				temp_resources = mdata['model_devi'][machine_idx]['resources']
# 				#assert isinstance(temp_machine, dict), "unsupported type of model_devi machine [%d]!" %machine_idx
# 				#assert isinstance(temp_resources, dict), "unsupported type of model_devi resources [%d]!"%machine_idx
# 				#assert temp_machine['machine_type'] == 'slurm', "Currently only support for Slurm!"
# 				temp_ssh_sess = SSHSession(temp_machine)
# 				cwd = os.getcwd()
# 				temp_context = SSHContext(cwd, temp_ssh_sess)
# 				if temp_machine['machine_type'] == 'lsf':
# 					temp_batch = LSF(temp_context)
# 				else:
# 					temp_batch = Slurm(temp_context)
# 				# For other type of machines, please add them using 'elif'.
# 				# Here slurm is selected as the final choice in convinience.
# 				command = temp_batch._make_squeue(temp_machine, temp_resources)
# 				ret, stdin, stdout, stderr = temp_batch.context.block_call(command)
# 				pd_response = stdout.read().decode('utf-8').split("\n")
# 				pd_count = len(pd_response)
# 				temp_context.clean()
# 				if pd_count ==0:
# 					mdata['model_devi_machine'] = temp_machine
# 					mdata['model_devi_resources'] = temp_resources
# 					mdata['model_devi_command'] = mdata['model_devi'][machine_idx]['command']
# 					mdata['model_devi_group_size'] =  mdata['model_devi'][machine_idx].get('group_size', 1)
# 					pd_flag = True
# 					break
# 				else:
# 					pd_count_list.append(pd_count)
# 			if not pd_flag:
# 				min_machine_idx = np.argsort(pd_count_list)[0]
# 				mdata['model_devi_machine'] = mdata['model_devi'][min_machine_idx]['machine']
# 				mdata['model_devi_resources'] = mdata['model_devi'][min_machine_idx]['resources']
# 				mdata['model_devi_command'] = mdata['model_devi'][min_machine_idx]['command']
# 				mdata['model_devi_group_size'] =  mdata['model_devi'][min_machine_idx].get('group_size', 1)
# 			with open("record.machine","w") as _outfile:
# 				profile = {}
# 				profile['purpose'] = 'model_devi'
# 				profile['machine'] = mdata['model_devi_machine']
# 				profile['resources'] = mdata['model_devi_resources']
# 				profile['group_size'] = mdata['model_devi_group_size']
# 				profile['command'] = mdata['model_devi_command']
#
# 				json.dump(profile, _outfile, indent = 4)
# 	return mdata
# def decide_fp_machine(mdata):
# 	if LooseVersion(mdata.get('api_version', '0.9')) >= LooseVersion('1.0'):
# 		mdata['fp_group_size'] = mdata['fp'][0]['resources']['group_size']
# 	if 'fp' in mdata:
# 		#ssert isinstance(mdata['fp']['machine'], list)
# 		#assert isinstance(mdata['fp']['resources'], list)
# 		#assert len(mdata['fp']['machine']) == len(mdata['fp']['resources'])
# 		continue_flag = False
# 		## decide whether to use an existing machine
# 		if 'record.machine' in os.listdir():
# 			try:
# 				with open('record.machine', 'r') as _infile:
# 					profile = json.load(_infile)
# 					if profile['purpose'] == 'fp':
# 						mdata['fp_machine'] = profile['machine']
# 						mdata['fp_resources'] = profile['resources']
# 						mdata['fp_command'] = profile['command']
# 						mdata['fp_group_size'] = profile['group_size']
#
# 						continue_flag = True
# 			except Exception:
# 				pass
# 		if ("hostname" not in mdata["fp"][0]["machine"]) or (len(mdata["fp"]) == 1):
# 			mdata["fp_machine"] = mdata["fp"][0]["machine"]
# 			mdata["fp_resources"] = mdata["fp"][0]["resources"]
# 			mdata["fp_command"] = mdata["fp"][0]["command"]
# 			#if "group_size" in mdata["train"][0]:
# 			mdata["fp_group_size"] = mdata["fp"][0].get("group_size", 1)
# 			continue_flag = True
#
#
# 		pd_count_list =[]
# 		pd_flag = False
# 		if not continue_flag:
# 			for machine_idx in range(len(mdata['fp'])):
# 				temp_machine = mdata['fp'][machine_idx]['machine']
# 				temp_resources = mdata['fp'][machine_idx]['resources']
# 				temp_ssh_sess = SSHSession(temp_machine)
# 				cwd = os.getcwd()
# 				temp_context = SSHContext(cwd, temp_ssh_sess)
# 				if temp_machine['machine_type'] == 'lsf':
# 					temp_batch = LSF(temp_context)
# 				else:
# 					temp_batch = Slurm(temp_context)
# 				# For other type of machines, please add them using 'elif'.
# 				# Here slurm is selected as the final choice in convinience.
# 				command = temp_batch._make_squeue(temp_machine, temp_resources)
# 				ret, stdin, stdout, stderr = temp_batch.context.block_call(command)
# 				pd_response = stdout.read().decode('utf-8').split("\n")
# 				pd_count = len(pd_response)
# 				temp_context.clean()
# 				#dlog.info(temp_machine["username"] + " " + temp_machine["hostname"] +  " " + str(pd_count))
# 				if pd_count ==0:
# 					mdata['fp_machine'] = temp_machine
# 					mdata['fp_resources'] = temp_resources
# 					mdata['fp_command'] = mdata['fp'][machine_idx]['command']
# 					mdata['fp_group_size'] =  mdata['fp'][machine_idx].get('group_size', 1)
# 					pd_flag = True
# 					break
# 				else:
# 					pd_count_list.append(pd_count)
# 			if not pd_flag:
# 				min_machine_idx = np.argsort(pd_count_list)[0]
# 				mdata['fp_machine'] = mdata['fp'][min_machine_idx]['machine']
# 				mdata['fp_resources'] = mdata['fp'][min_machine_idx]['resources']
# 				mdata['fp_command'] = mdata['fp'][min_machine_idx]['command']
# 				mdata['fp_group_size'] =  mdata['fp'][min_machine_idx].get('group_size',1)
#
# 			with open("record.machine","w") as _outfile:
# 					profile = {}
# 					profile['purpose'] = 'fp'
# 					profile['machine'] = mdata['fp_machine']
# 					profile['resources'] = mdata['fp_resources']
# 					profile['group_size'] = mdata['fp_group_size']
# 					profile['command'] = mdata['fp_command']
# 					json.dump(profile, _outfile, indent = 4)
# 	return mdata




import os, sys, json, glob, shutil
import dpdata
import numpy as np
from monty.serialization import loadfn, dumpfn
import unittest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__package__ = 'auto_test'

from .context import make_kspacing_kpoints
from .context import setUpModule

from pymatgen.io.vasp import Incar
from dpgen.auto_test.common_equi import make_equi, post_equi
from dpgen.auto_test.calculator import make_calculator


class TestEqui(unittest.TestCase):
    jdata = {
        "structures": ["confs/std-fcc"],
        "interaction": {
            "type": "vasp",
            "incar": "vasp_input/INCAR.rlx",
            "potcar_prefix": ".",
            "potcars": {"Al": "vasp_input/POT_Al"}
        },
        "relaxation": {
            "cal_type": "relaxation",
            "cal_setting":{"input_prop": "vasp_input/INCAR"}
        }
    }

    def tearDown(self):
        if os.path.isfile('confs/std-fcc/POSCAR'):
           os.remove('confs/std-fcc/POSCAR')
        if os.path.exists('confs/std-fcc/relaxation'):
            shutil.rmtree('confs/std-fcc/relaxation')

    def test_make_equi(self):
        confs = self.jdata["structures"]
        inter_param = self.jdata["interaction"]
        relax_param = self.jdata["relaxation"]
        make_equi(confs, inter_param, relax_param)

        self.assertTrue(os.path.isfile("confs/std-fcc/POSCAR"))

        target_path = 'confs/std-fcc/relaxation/relax_task'
        source_path = 'vasp_input'

        incar0 = Incar.from_file(os.path.join('vasp_input', 'INCAR'))
        incar1 = Incar.from_file(os.path.join(target_path, 'INCAR'))
        self.assertTrue(incar0 == incar1)

        with open(os.path.join('vasp_input', 'POT_Al')) as fp:
            pot0 = fp.read()
        with open(os.path.join(target_path, 'POTCAR')) as fp:
            pot1 = fp.read()
        self.assertEqual(pot0, pot1)

        self.assertTrue(os.path.isfile(os.path.join(target_path, 'KPOINTS')))

        task_json_file = os.path.join(target_path, 'task.json')
        self.assertTrue(os.path.isfile(task_json_file))
        task_json = loadfn(task_json_file)
        self.assertEqual(task_json, relax_param)

        inter_json_file = os.path.join(target_path, 'inter.json')
        self.assertTrue(os.path.isfile(inter_json_file))
        inter_json = loadfn(inter_json_file)
        self.assertEqual(inter_json, inter_param)

        self.assertTrue(os.path.islink(os.path.join(target_path, 'POSCAR')))




"""This module ensures input in the examples directory
could pass the argument checking.
"""
import unittest
import json
from pathlib import Path

from dpgen.util import normalize
from dpgen.data.arginfo import (
    init_reaction_jdata_arginfo,
)
from dpgen.simplify.arginfo import (
    simplify_jdata_arginfo,
)

init_reaction_jdata = init_reaction_jdata_arginfo()
simplify_jdata = simplify_jdata_arginfo()

# directory of examples
p_examples = Path(__file__).parent.parent / "examples"

# input_files : tuple[tuple[Argument, Path]]
#   tuple of example list
input_files = (
    (init_reaction_jdata, p_examples / "init" / "reaction.json"),
    (simplify_jdata, p_examples / "simplify" / "qm7.json"),
)


class TestExamples(unittest.TestCase):
    def test_arguments(self):
        for arginfo, fn in input_files:
            fn = str(fn)
            with self.subTest(fn=fn):
                with open(fn) as f:
                    data = json.load(f)
                normalize(arginfo, data)



from typing import Dict, List
from dargs import Argument, Variant

from dpgen.arginfo import general_mdata_arginfo


def run_mdata_arginfo() -> Argument:
    """Generate arginfo for dpgen run mdata.
    
    Returns
    -------
    Argument
        arginfo
    """
    return general_mdata_arginfo("run_mdata", ("train", "model_devi", "fp"))

# basics
def basic_args() -> List[Argument]:
    doc_type_map = 'Atom types.'
    doc_mass_map = 'Standard atom weights.'
    doc_use_ele_temp = 'Currently only support fp_style vasp. \n\n\
- 0: no electron temperature. \n\n\
- 1: eletron temperature as frame parameter. \n\n\
- 2: electron temperature as atom parameter.'

    return [
        Argument("type_map", list, optional=False, doc=doc_type_map),
        Argument("mass_map", list, optional=False, doc=doc_mass_map),
        Argument("use_ele_temp", int, optional=True,
                 default=0, doc=doc_use_ele_temp),
    ]


def data_args() -> List[Argument]:
    doc_init_data_prefix = 'Prefix of initial data directories.'
    doc_init_data_sys = 'Directories of initial data. You may use either absolute or relative path here. Systems will be detected recursively in the directories.'
    doc_sys_format = 'Format of initial data.'
    doc_init_batch_size = 'Each number is the batch_size of corresponding system for training in init_data_sys. One recommended rule for setting the sys_batch_size and init_batch_size is that batch_size mutiply number of atoms ot the stucture should be larger than 32. If set to auto, batch size will be 32 divided by number of atoms.'
    doc_sys_configs_prefix = 'Prefix of sys_configs.'
    doc_sys_configs = 'Containing directories of structures to be explored in iterations.Wildcard characters are supported here.'
    doc_sys_batch_size = 'Each number is the batch_size for training of corresponding system in sys_configs. If set to auto, batch size will be 32 divided by number of atoms.'

    return [
        Argument("init_data_prefix", str, optional=True,
                 doc=doc_init_data_prefix),
        Argument("init_data_sys", list,
                 optional=False, doc=doc_init_data_sys),
        Argument("sys_format", str, optional=True, default='vasp/poscar', doc=doc_sys_format),
        Argument("init_batch_size", str, optional=True,
                 doc=doc_init_batch_size),
        Argument("sys_configs_prefix", str, optional=True,
                 doc=doc_sys_configs_prefix),
        Argument("sys_configs", list, optional=False, doc=doc_sys_configs),
        Argument("sys_batch_size", list, optional=True,
                 doc=doc_sys_batch_size),
    ]

# Training


def training_args() -> List[Argument]:
    doc_numb_models = 'Number of models to be trained in 00.train. 4 is recommend.'
    doc_training_iter0_model_path = 'The model used to init the first iter training. Number of element should be equal to numb_models.'
    doc_training_init_model = 'Iteration > 0, the model parameters will be initilized from the model trained at the previous iteration. Iteration == 0, the model parameters will be initialized from training_iter0_model_path.'
    doc_default_training_param = 'Training parameters for deepmd-kit in 00.train. You can find instructions from here: (https://github.com/deepmodeling/deepmd-kit).'
    doc_dp_compress = 'Use dp compress to compress the model.'

    return [
        Argument("numb_models", int, optional=False, doc=doc_numb_models),
        Argument("training_iter0_model_path", list, optional=True,
                 doc=doc_training_iter0_model_path),
        Argument("training_init_model", bool, optional=True,
                 doc=doc_training_init_model),
        Argument("default_training_param", dict, optional=False,
                 doc=doc_default_training_param),
        Argument("dp_compress", bool, optional=True,
                 default=False, doc=doc_dp_compress),
    ]


# Exploration
def model_devi_jobs_args() -> List[Argument]:
    # this may be not correct
    doc_sys_idx = 'Systems to be selected as the initial structure of MD and be explored. The index corresponds exactly to the sys_configs.'
    doc_temps = 'Temperature (K) in MD.'
    doc_press = 'Pressure (Bar) in MD.'
    doc_trj_freq = 'Frequecy of trajectory saved in MD.'
    doc_nsteps = 'Running steps of MD.'
    doc_ensembles = 'Determining which ensemble used in MD, options include “npt” and “nvt”.'
    doc_neidelay = 'delay building until this many steps since last build.'
    doc_taut = 'Coupling time of thermostat (ps).'
    doc_taup = 'Coupling time of barostat (ps).'
    doc_model_devi_f_trust_lo = 'Lower bound of forces for the selection. If dict, should be set for each index in sys_idx, respectively.'
    doc_model_devi_f_trust_hi = 'Upper bound of forces for the selection. If dict, should be set for each index in sys_idx, respectively.'
    doc_model_devi_v_trust_lo = 'Lower bound of virial for the selection. If dict, should be set for each index in sys_idx, respectively. Should be used with DeePMD-kit v2.x.'
    doc_model_devi_v_trust_hi = 'Upper bound of virial for the selection. If dict, should be set for each index in sys_idx, respectively. Should be used with DeePMD-kit v2.x.'

    args = [
        Argument("sys_idx", list, optional=False, doc=doc_sys_idx),
        Argument("temps", list, optional=False, doc=doc_temps),
        Argument("press", list, optional=False, doc=doc_press),
        Argument("trj_freq", int, optional=False, doc=doc_trj_freq),
        Argument("nsteps", int, optional=False, doc=doc_nsteps),
        Argument("ensembles", str, optional=False, doc=doc_ensembles),
        Argument("neidelay", int, optional=True, doc=doc_neidelay),
        Argument("taut", float, optional=True, doc=doc_taut),
        Argument("taup", float, optional=True, doc=doc_taup),
        Argument("model_devi_f_trust_lo", [
                 float, dict], optional=False, doc=doc_model_devi_f_trust_lo),
        Argument("model_devi_f_trust_hi", [
                 float, dict], optional=False, doc=doc_model_devi_f_trust_hi),
        Argument("model_devi_v_trust_lo", [
                 float, dict], optional=False, doc=doc_model_devi_v_trust_lo),
        Argument("model_devi_v_trust_hi", [
                 float, dict], optional=False, doc=doc_model_devi_v_trust_hi),
    ]

    doc_model_devi_jobs = 'Settings for exploration in 01.model_devi. Each dict in the list corresponds to one iteration. The index of model_devi_jobs exactly accord with index of iterations'
    return Argument("model_devi_jobs", list, args, [], repeat=True, doc=doc_model_devi_jobs)


def model_devi_lmp_args() -> List[Argument]:
    doc_model_devi_dt = 'Timestep for MD. 0.002 is recommend.'
    doc_model_devi_skip = 'Number of structures skipped for fp in each MD.'
    doc_model_devi_f_trust_lo = 'Lower bound of forces for the selection. If list or dict, should be set for each index in sys_configs, respectively.'
    doc_model_devi_f_trust_hi = 'Upper bound of forces for the selection. If list or dict, should be set for each index in sys_configs, respectively.'
    doc_model_devi_v_trust_lo = 'Lower bound of virial for the selection. If list or dict, should be set for each index in sys_configs, respectively. Should be used with DeePMD-kit v2.x.'
    doc_model_devi_v_trust_hi = 'Upper bound of virial for the selection. If list or dict, should be set for each index in sys_configs, respectively. Should be used with DeePMD-kit v2.x.'
    doc_model_devi_adapt_trust_lo = 'Adaptively determines the lower trust levels of force and virial. This option should be used together with model_devi_numb_candi_f, model_devi_numb_candi_v and optionally with model_devi_perc_candi_f and model_devi_perc_candi_v. dpgen will make two sets:\n\n\
- 1. From the frames with force model deviation lower than model_devi_f_trust_hi, select max(model_devi_numb_candi_f, model_devi_perc_candi_f*n_frames) frames with largest force model deviation. \n\n\
- 2. From the frames with virial model deviation lower than model_devi_v_trust_hi, select max(model_devi_numb_candi_v, model_devi_perc_candi_v*n_frames) frames with largest virial model deviation. \n\n\
The union of the two sets is made as candidate dataset.'
    doc_model_devi_numb_candi_f = 'See model_devi_adapt_trust_lo.'
    doc_model_devi_numb_candi_v = 'See model_devi_adapt_trust_lo.'
    doc_model_devi_perc_candi_f = 'See model_devi_adapt_trust_lo.'
    doc_model_devi_perc_candi_v = 'See model_devi_adapt_trust_lo.'
    doc_model_devi_f_avg_relative = 'Normalized the force model deviations by the RMS force magnitude along the trajectory. This key should not be used with use_relative.'
    doc_model_devi_clean_traj = 'If type of model_devi_clean_traj is bool type then it denote whether to clean traj folders in MD since they are too large. If it is Int type, then the most recent n iterations of traj folders will be retained, others will be removed.'
    doc_model_devi_nopbc = 'Assume open boundary condition in MD simulations.'
    doc_model_devi_activation_func = 'Set activation functions for models, length of the list should be the same as numb_models, and two elements in the list of string respectively assign activation functions to the embedding and fitting nets within each model. Backward compatibility: the orginal "list of String" format is still supported, where embedding and fitting nets of one model use the same activation function, and the length of the list should be the same as numb_models.'

    return [
        model_devi_jobs_args(),
        Argument("model_devi_dt", float,
                 optional=False, doc=doc_model_devi_dt),
        Argument("model_devi_skip", int, optional=False,
                 doc=doc_model_devi_skip),
        Argument("model_devi_f_trust_lo", [
                 float, list, dict], optional=False, doc=doc_model_devi_f_trust_lo),
        Argument("model_devi_f_trust_hi", [
                 float, list, dict], optional=False, doc=doc_model_devi_f_trust_hi),
        Argument("model_devi_v_trust_lo", [
                 float, list, dict], optional=False, doc=doc_model_devi_v_trust_lo),
        Argument("model_devi_v_trust_hi", [
                 float, list, dict], optional=False, doc=doc_model_devi_v_trust_hi),
        Argument("model_devi_adapt_trust_lo", bool, optional=True,
                 doc=doc_model_devi_adapt_trust_lo),
        Argument("model_devi_numb_candi_f", int, optional=True,
                 doc=doc_model_devi_numb_candi_f),
        Argument("model_devi_numb_candi_v", int, optional=True,
                 doc=doc_model_devi_numb_candi_v),
        Argument("model_devi_perc_candi_f", float,
                 optional=True, doc=doc_model_devi_perc_candi_f),
        Argument("model_devi_perc_candi_v", float,
                 optional=True, doc=doc_model_devi_perc_candi_v),
        Argument("model_devi_f_avg_relative", bool, optional=True,
                 doc=doc_model_devi_f_avg_relative),
        Argument("model_devi_clean_traj", [
                 bool, int], optional=False, doc=doc_model_devi_clean_traj),
        Argument("model_devi_nopbc", bool, optional=False,
                 doc=doc_model_devi_nopbc),
        Argument("model_devi_activation_func", list, optional=True,
                 doc=doc_model_devi_activation_func),
    ]


def model_devi_args() -> List[Variant]:
    doc_model_devi_engine = "Engine for the model deviation task."
    return [Variant("model_devi_engine", [
            Argument("lammps", dict, model_devi_lmp_args(), doc="LAMMPS"),
        ], default_tag="lammps", optional=True, doc=doc_model_devi_engine)]


# Labeling
# vasp
def fp_style_vasp_args() -> List[Argument]:
    doc_fp_pp_path = 'Directory of psuedo-potential file to be used for 02.fp exists.'
    doc_fp_pp_files = 'Psuedo-potential file to be used for 02.fp. Note that the order of elements should correspond to the order in type_map.'
    doc_fp_incar = 'Input file for VASP. INCAR must specify KSPACING and KGAMMA.'
    doc_fp_aniso_kspacing = 'Set anisotropic kspacing. Usually useful for 1-D or 2-D materials. Only support VASP. If it is setting the KSPACING key in INCAR will be ignored.'
    doc_cvasp = 'If cvasp is true, DP-GEN will use Custodian to help control VASP calculation.'

    return [
        Argument("fp_pp_path", str, optional=False, doc=doc_fp_pp_path),
        Argument("fp_pp_files", list, optional=False, doc=doc_fp_pp_files),
        Argument("fp_incar", str, optional=False, doc=doc_fp_incar),
        Argument("fp_aniso_kspacing", list, optional=False,
                 doc=doc_fp_aniso_kspacing),
        Argument("cvasp", bool, optional=False, doc=doc_cvasp),
    ]


# gaussian
def fp_style_gaussian_args() -> List[Argument]:
    doc_keywords = 'Keywords for Gaussian input.'
    doc_multiplicity = 'Spin multiplicity for Gaussian input. If set to auto, the spin multiplicity will be detected automatically. If set to frag, the "fragment=N" method will be used.'
    doc_nproc = 'The number of processors for Gaussian input.'

    args = [
        Argument("keywords", [str or list],
                 optional=False, doc=doc_keywords),
        Argument("multiplicity", [int or str],
                 optional=False, doc=doc_multiplicity),
        Argument("nproc", int, optional=False, doc=doc_nproc),
    ]

    doc_use_clusters = 'If set to true, clusters will be taken instead of the whole system. This option does not work with DeePMD-kit 0.x.'
    doc_cluster_cutoff = 'The cutoff radius of clusters if use_clusters is set to true.'
    doc_fp_params_gaussian = 'Parameters for Gaussian calculation.'

    return [
        Argument("use_clusters", bool, optional=True, default=False, doc=doc_use_clusters),
        Argument("cluster_cutoff", float,
                 optional=True, doc=doc_cluster_cutoff),
        Argument("fp_params", dict, args, [],
                 optional=False, doc=doc_fp_params_gaussian),
    ]

# siesta
def fp_style_siesta_args() -> List[Argument]:
    doc_ecut = 'Define the plane wave cutoff for grid.'
    doc_ediff = 'Tolerance of Density Matrix.'
    doc_kspacing = 'Sample factor in Brillouin zones.'
    doc_mixingweight = 'Proportion a of output Density Matrix to be used for the input Density Matrix of next SCF cycle (linear mixing).'
    doc_NumberPulay = 'Controls the Pulay convergence accelerator.'

    args = [
        Argument("ecut", int, optional=False, doc=doc_ecut),
        Argument("ediff", float, optional=False, doc=doc_ediff),
        Argument("kspacing", float, optional=False, doc=doc_kspacing),
        Argument("mixingweight", float, optional=False, doc=doc_mixingweight),
        Argument("NumberPulay", int, optional=False, doc=doc_NumberPulay),
    ]

    doc_use_clusters = 'If set to true, clusters will be taken instead of the whole system. This option does not work with DeePMD-kit 0.x.'
    doc_cluster_cutoff = 'The cutoff radius of clusters if use_clusters is set to true.'
    doc_fp_params_siesta = 'Parameters for siesta calculation.'

    return [
        Argument("use_clusters", bool, optional=False, doc=doc_use_clusters),
        Argument("cluster_cutoff", float,
                 optional=False, doc=doc_cluster_cutoff),
        Argument("fp_params", dict, args, [],
                 optional=False, doc=doc_fp_params_siesta),
    ]

# cp2k


def fp_style_cp2k_args() -> List[Argument]:
    doc_user_fp_params = 'Parameters for cp2k calculation. find detail in manual.cp2k.org. only the kind section must be set before use. we assume that you have basic knowledge for cp2k input.'
    doc_external_input_path = 'Conflict with key:user_fp_params, use the template input provided by user, some rules should be followed, read the following text in detail.'

    return [
        Argument("user_fp_params", dict, optional=False,
                 doc=doc_user_fp_params),
        Argument("external_input_path", str, optional=False,
                 doc=doc_external_input_path),
    ]


def fp_style_variant_type_args() -> Variant:
    doc_fp_style = 'Software for First Principles.'

    return Variant("fp_style", [Argument("vasp", dict, fp_style_vasp_args()),
                                Argument("gaussian", dict,
                                         fp_style_gaussian_args()),
                                Argument("siesta", dict,
                                         fp_style_siesta_args()),
                                Argument("cp2k", dict, fp_style_cp2k_args())],
                   optional=False,
                   doc=doc_fp_style)


def fp_args() -> List[Argument]:
    doc_fp_task_max = 'Maximum of structures to be calculated in 02.fp of each iteration.'
    doc_fp_task_min = 'Minimum of structures to be calculated in 02.fp of each iteration.'
    doc_fp_accurate_threshold = 'If the accurate ratio is larger than this number, no fp calculation will be performed, i.e. fp_task_max = 0.'
    doc_fp_accurate_soft_threshold = 'If the accurate ratio is between this number and fp_accurate_threshold, the fp_task_max linearly decays to zero.'
    doc_fp_cluster_vacuum = 'If the vacuum size is smaller than this value, this cluster will not be choosen for labeling.'

    return [
        Argument("fp_task_max", int, optional=False, doc=doc_fp_task_max),
        Argument("fp_task_min", int, optional=False, doc=doc_fp_task_min),
        Argument("fp_accurate_threshold", float,
                 optional=True, doc=doc_fp_accurate_threshold),
        Argument("fp_accurate_soft_threshold", float,
                 optional=True, doc=doc_fp_accurate_soft_threshold),
        Argument("fp_cluster_vacuum", float,
                 optional=True, doc=doc_fp_cluster_vacuum),
    ]


def run_jdata_arginfo() -> Argument:
    """Argument information for dpgen run mdata.
    
    Returns
    -------
    Argument
        argument information
    """
    doc_run_jdata = "param.json file"
    return Argument("run_jdata",
                    dict,
                    sub_fields=basic_args() + data_args() + training_args() + fp_args(),
                    sub_variants=model_devi_args() + [fp_style_variant_type_args()],
                    doc=doc_run_jdata)
